from urllib.parse import urlparse, urlunparse, urlencode, parse_qsl, ParseResult

def set_url_query_param(url: str, param_name: str, param_value: str):
    """
    Given a url, set or replace a query parameter and return the modified url.

    >> set_url_query_param('https://api.authorizon.com/opal/data/config', 'token', 'secret')
    'https://api.authorizon.com/opal/data/config?token=secret'

    >> set_url_query_param('https://api.authorizon.com/opal/data/config&some=var', 'token', 'secret')
    'https://api.authorizon.com/opal/data/config&some=var?token=secret'
    """
    parsed_url: ParseResult = urlparse(url)

    query_params: dict = dict(parse_qsl(parsed_url.query))
    query_params[param_name] = param_value
    new_query_string = urlencode(query_params)

    return urlunparse((
        parsed_url.scheme,
        parsed_url.netloc,
        parsed_url.path,
        parsed_url.params,
        new_query_string,
        parsed_url.fragment,
    ))




from typing import Callable, Dict, List
import typer

from typer.main import Typer
from .types import ConfiEntry
import click


def create_click_cli(confi_entries:Dict[str,ConfiEntry], callback:Callable):
    cli = callback
    for key, entry in confi_entries.items():
        option_kwargs = entry.get_cli_option_kwargs()
        # make the key fit cmd-style (i.e. kebab-case)
        adjusted_key = entry.key.lower().replace("_","-")
        keys = [f"--{adjusted_key}", entry.key]
        # add flag if given (i.e '-t' option)
        if entry.flags is not None:
            keys.extend(entry.flags)
        # use lower case as the key, and as is (no prefix, and no case altering) as the name
        # see https://click.palletsprojects.com/en/7.x/options/#name-your-options
        cli = click.option(*keys, **option_kwargs)(cli)
    # pass context
    cli = click.pass_context(cli)
    # wrap in group
    cli = click.group(invoke_without_command=True)(cli)
    return cli

def get_cli_object_for_config_objects(config_objects:list, typer_app:Typer=None, help:str=None, on_start:Callable=None):  
    # callback to save CLI results back to objects   
    def callback(ctx, **kwargs):
        if callable(on_start):
            on_start(ctx, **kwargs)

        for key, value in kwargs.items():
            # find the confi-object which the key belongs to and ...
            for config_obj in config_objects:
                if key in config_obj.entries:
                    # ... update that object with the new value
                    setattr(config_obj,key, value) 
                    config_obj._entries[key].value = value

    if help is not None:
        callback.__doc__ = help
    # Create a merged config-entires map
    entries = {}
    for config_obj in config_objects:
        entries.update(config_obj.entries)
    # convert to a click-cli group 
    click_group = create_click_cli(entries, callback)
    # add the typer app into our click group
    if typer_app is not None:
        typer_click_object = typer.main.get_command(typer_app)
        # add the app commands directly to out click app
        for name, cmd in typer_click_object.commands.items():
            click_group.add_command(cmd, name)
    return click_group    




class MainTexts:

    def __init__(self, first_line, name):

        self.header = f"""\b
    {first_line}
    Open-Policy Administration Layer - {name}\b\f"""

        self.docs = f"""\b
    Config top level options:
        - Use env-vars (same as cmd options) but uppercase 
            and with "_" instead of "-"; all prefixed with "OPAL_"
        - Use command line options as detailed by '--help'
        - Use .env or .ini files    

    \b
    Examples:
        - opal-{name} --help                           Detailed help on CLI
        - opal-{name} run --help                       Help on run command
        - opal-{name} run --engine-type gunicorn       Run {name} with gunicorn
    \b
    """



import os
from typing import Optional

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization

from opal_common.authentication.types import EncryptionKeyFormat, PrivateKey, PublicKey


def to_bytes(key: str, encoding: str = 'utf-8'):
    """
    crypto lib expect 'bytes' keys, convert 'str' keys to 'bytes'
    """
    return key.encode(encoding)

def maybe_decode_multiline_key(key: str) -> bytes:
    """
    if key contents are passed via env var, we allow to encode multiline keys
    with a simple replace of each newline (\n) char with underscore (_).

    this method detects if the provided key contains such encoding, and if so reverses it.
    """
    if "\n" not in key:
        key = key.replace("_", "\n")
        if not key.endswith("\n"):
            key = key + "\n"
    return to_bytes(key)


def cast_private_key(value: str, key_format: EncryptionKeyFormat, passphrase: Optional[str] = None) -> Optional[PrivateKey]:
    """
    Parse a string into a valid cryptographic private key.
    the string can represent a file path in which the key exists, or the actual key contents.
    """
    if value is None:
        return None

    if isinstance(value, PrivateKey.__args__):
        return value

    if passphrase is None:
        password = None
    else:
        password = passphrase.encode('utf-8')

    key_path = os.path.expanduser(value)
    if os.path.isfile(key_path):
        raw_key = open(key_path, "rb").read()
    else:
        raw_key = maybe_decode_multiline_key(value)

    if key_format == EncryptionKeyFormat.pem:
        return serialization.load_pem_private_key(raw_key, password=password, backend=default_backend())

    if key_format == EncryptionKeyFormat.ssh:
        return serialization.load_ssh_private_key(raw_key, password=password, backend=default_backend())

    if key_format == EncryptionKeyFormat.der:
        return serialization.load_der_private_key(raw_key, password=password, backend=default_backend())


def cast_public_key(value: str, key_format: EncryptionKeyFormat) -> Optional[PublicKey]:
    """
    Parse a string into a valid cryptographic public key.
    the string can represent a file path in which the key exists, or the actual key contents.
    """
    if value is None:
        return None

    if isinstance(value, PublicKey.__args__):
        return value

    key_path = os.path.expanduser(value)
    if os.path.isfile(key_path):
        raw_key = open(key_path, "rb").read()
    elif key_format == EncryptionKeyFormat.ssh: # ssh key format is one line
        raw_key = to_bytes(value)
    else:
        raw_key = maybe_decode_multiline_key(value)

    if key_format == EncryptionKeyFormat.pem:
        return serialization.load_pem_public_key(raw_key, backend=default_backend())

    if key_format == EncryptionKeyFormat.ssh:
        return serialization.load_ssh_public_key(raw_key, backend=default_backend())

    if key_format == EncryptionKeyFormat.der:
        return serialization.load_der_public_key(raw_key, backend=default_backend())



from typing import List
from pathlib import Path

from opal_common.schemas.policy import DataModule, PolicyBundle, RegoModule


class BundleUtils:
    MAX_INDEX = 10000

    @staticmethod
    def sorted_policy_modules_to_load(bundle: PolicyBundle) -> List[RegoModule]:
        """
        policy modules sorted according to manifest
        """
        manifest_paths = [Path(path) for path in bundle.manifest]
        def key_function(module: RegoModule) -> int:
            """
            this method reduces the module to a number that can be act as sorting key.

            the number is the index in the manifest list, so basically we sort according to manifest.
            """
            try:
                return manifest_paths.index(Path(module.path))
            except ValueError:
                return BundleUtils.MAX_INDEX
        return sorted(bundle.policy_modules, key=key_function)

    @staticmethod
    def sorted_data_modules_to_load(bundle: PolicyBundle) -> List[DataModule]:
        """
        data modules sorted according to manifest
        """
        manifest_paths = [Path(path) for path in bundle.manifest]
        def key_function(module: DataModule) -> int:
            try:
                return manifest_paths.index(Path(module.path))
            except ValueError:
                return BundleUtils.MAX_INDEX
        return sorted(bundle.data_modules, key=key_function)

    @staticmethod
    def sorted_policy_modules_to_delete(bundle: PolicyBundle) -> List[Path]:
        # already sorted
        return bundle.deleted_files.policy_modules

    @staticmethod
    def sorted_data_modules_to_delete(bundle: PolicyBundle) -> List[Path]:
        # already sorted
        return bundle.deleted_files.data_modules


import os
import asyncio
from functools import partial
from typing import Callable, Coroutine, List, Union

from git.objects.commit import Commit
from opal_common.logger import logger

OnNewPolicyCallback = Callable[[
    Commit, Commit], Coroutine]
OnPolicyFailureCallback = Callable[[Exception], Coroutine]


class BasePolicySource:
    """
    Base class to support git and api policy source

    Args:
        remote_source_url(str): the base address to request the policy from
        local_clone_path(str):  path for the local git to manage policies
        polling_interval(int):  how many seconds need to wait between polling
    """

    def __init__(
        self,
        remote_source_url: str,
        local_clone_path: str,
        polling_interval: int = 0,
    ):
        self._on_failure_callbacks: List[OnNewPolicyCallback] = []
        self._on_new_policy_callbacks: List[OnPolicyFailureCallback] = []
        self._polling_interval = polling_interval
        self._polling_task = None
        self.remote_source_url = remote_source_url
        self.local_clone_path = os.path.expanduser(local_clone_path)

    def add_on_new_policy_callback(self, callback: OnNewPolicyCallback):
        """
        Register a callback that will be called when new policy
        are detected on the monitored repo (after a pull).
        """
        self._on_new_policy_callbacks.append(callback)

    def add_on_failure_callback(self, callback: OnPolicyFailureCallback):
        """
        Register a callback that will be called when failure occurred.
        """
        self._on_failure_callbacks.append(callback)

    async def get_initial_policy_state_from_remote(self):
        """
        init remote data to local repo
        """
        raise NotImplementedError()

    async def check_for_changes(self):
        """
        trigger check for policy change
        """
        raise NotImplementedError()

    async def run(self):
        """
        potentially starts the polling task
        """
        await self.get_initial_policy_state_from_remote()

        if (self._polling_interval > 0):
            logger.info(
                "Launching polling task, interval: {interval} seconds", interval=self._polling_interval)
            self._start_polling_task(self.check_for_changes)
        else:
            logger.info("Polling task is off")

    async def stop(self):
        return await self._stop_polling_task()

    def _start_polling_task(self, polling_task):
        if self._polling_task is None and self._polling_interval > 0:
            self._polling_task = asyncio.create_task(self._do_polling(polling_task))

    async def _do_polling(self, polling_task):
        """
        optional task to periodically check the remote for changes (git pull and compare hash).
        """
        while True:
            await polling_task()
            await asyncio.sleep(self._polling_interval)

    async def _stop_polling_task(self):
        if self._polling_task is not None:
            self._polling_task.cancel()
            try:
                await self._polling_task
            except asyncio.CancelledError:
                pass

    async def _on_new_policy(self, old: Commit, new: Commit):
        """
        triggers callbacks registered with on_new_policy().
        """
        await self._run_callbacks(self._on_new_policy_callbacks, old, new)

    async def _on_failed(self, exc: Exception):
        """
        will be triggered if a failure occurred.
        triggers callbacks registered with on_git_failed().
        """
        await self._run_callbacks(self._on_failure_callbacks, exc)

    async def _run_callbacks(self, handlers, *args, **kwargs):
        """
        triggers a list of callbacks
        """
        await asyncio.gather(*(callback(*args, **kwargs) for callback in handlers))

    async def _on_git_failed(self, exc: Exception):
        """
        will be triggered if a git failure occurred (i.e: repo does not exist, can't clone, etc).
        triggers callbacks registered with on_git_failed().
        """
        await self._run_callbacks(self._on_failure_callbacks, exc)



import os
import uuid
import shutil
import asyncio

from functools import partial
from typing import Optional, Generator
from pathlib import Path
from tenacity import retry, wait, stop, RetryError
from git import Repo, GitError, GitCommandError

from opal_common.logger import logger
from opal_common.git.exceptions import GitFailed
from opal_common.config import opal_common_config
from opal_common.utils import get_filepaths_with_glob


SSH_PREFIX = "ssh://"
GIT_SSH_USER_PREFIX = "git@"

def is_ssh_repo_url(repo_url: str):
    """
    return True if the repo url uses SSH authentication.
    (see: https://docs.github.com/en/github/authenticating-to-github/connecting-to-github-with-ssh)
    """
    return repo_url.startswith(SSH_PREFIX) or repo_url.startswith(GIT_SSH_USER_PREFIX)


class CloneResult:
    """
    wraps a git.Repo instance but knows if the repo was initialized with a url
    and cloned from a remote repo, or was initialed from a local `.git` repo.
    """
    def __init__(self, repo: Repo):
        self._repo = repo

    @property
    def repo(self) -> Repo:
        """
        the wrapped repo instance
        """
        return self._repo


class RepoClonePathFinder:
    """
    We are cloning the policy repo into a unique random subdirectory of a base path.

    This class knows how to such clones, so we can discard previous ones, but also so
    that siblings workers (who are not the master who decided where to clone) can also
    find the current clone by globing the base dir.
    """
    def __init__(self, base_clone_path: str, clone_subdirectory_prefix: str):
        if not base_clone_path:
            raise ValueError("base_clone_path cannot be empty!")

        if not clone_subdirectory_prefix:
            raise ValueError("clone_subdirectory_prefix cannot be empty!")

        self._base_clone_path = os.path.expanduser(base_clone_path)
        self._clone_subdirectory_prefix = clone_subdirectory_prefix

    def get_clone_subdirectories(self) -> Generator[str, None, None]:
        """
        a generator yielding all the subdirectories of the base clone path
        that are matching the clone pattern.

        Yields:
            the next subdirectory matching the pattern
        """
        folders_with_pattern = get_filepaths_with_glob(self._base_clone_path, f"{self._clone_subdirectory_prefix}*")
        for folder in folders_with_pattern:
            yield folder

    def get_single_clone_path(self) -> Optional[str]:
        """
        searches for the single clone subdirectory in existance.

        If found no such subdirectory or if found more than one (multiple matching subdirectories) - will return None.
        otherwise: will return the single and only clone.
        """
        subdirectories = list(self.get_clone_subdirectories())
        if len(subdirectories) != 1:
            return None
        return subdirectories[0]

    def create_new_clone_path(self) -> str:
        folder_name = f"{self._clone_subdirectory_prefix}-{uuid.uuid4().hex}"
        full_local_repo_path = os.path.join(self._base_clone_path, folder_name)
        os.makedirs(full_local_repo_path, exist_ok=True)
        return full_local_repo_path

class RepoCloner:
    """
    simple wrapper for git.Repo() to simplify other classes that need to deal
    with the case where a repo must be cloned from url *only if* the repo does
    not already exists locally, and otherwise initialize the repo instance from
    the repo already existing on the filesystem.
    """

    # wait indefinitely until successful
    DEFAULT_RETRY_CONFIG = {
        'wait': wait.wait_random_exponential(multiplier=0.5, max=30),
    }

    def __init__(
        self,
        repo_url: str,
        clone_path: str,
        branch_name: str = "master",
        retry_config = None,
        ssh_key: Optional[str] = None,
        ssh_key_file_path: Optional[str] = None,
        clone_timeout: int = 0,
    ):
        """inits the repo cloner.

        Args:
            repo_url (str): the url to the remote repo we want to clone
            clone_path (str): the target local path in our file system we want the
                repo to be cloned to
            retry_config (dict): Tenacity.retry config (@see https://tenacity.readthedocs.io/en/latest/api.html#retry-main-api)
            ssh_key (str, optional): private ssh key used to gain access to the cloned repo
            ssh_key_file_path (str, optional): local path to save the private ssh key contents
        """
        if repo_url is None:
            raise ValueError("must provide repo url!")


        self.url = repo_url
        self.path = os.path.expanduser(clone_path)
        self.branch_name = branch_name
        self._ssh_key = ssh_key
        self._ssh_key_file_path = ssh_key_file_path or opal_common_config.GIT_SSH_KEY_FILE
        self._retry_config = retry_config if retry_config is not None else self.DEFAULT_RETRY_CONFIG
        if clone_timeout > 0:
            self._retry_config.update({'stop': stop.stop_after_delay(clone_timeout)})

    async def clone(self) -> CloneResult:
        """
        initializes a git.Repo and returns the clone result.
        it either:
            - does not found a cloned repo locally and clones from remote url
            - finds a cloned repo locally and does not clone from remote.
        """
        logger.info("Cloning repo from '{url}' to '{to_path}' (branch: '{branch}')", url=self.url, to_path=self.path, branch=self.branch_name)
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, self._attempt_clone_from_url)

    def _attempt_clone_from_url(self) -> CloneResult:
        """
        clones the repo from url or throws GitFailed
        """
        env = self._provide_git_ssh_environment()
        _clone_func = partial(self._clone, env=env)
        _clone_with_retries = retry(**self._retry_config)(_clone_func)
        try:
            repo: Repo = _clone_with_retries()
        except (GitError, GitCommandError) as e:
            raise GitFailed(e)
        except RetryError as e:
            logger.exception("cannot clone policy repo: {error}", error=e)
            raise GitFailed(e)
        else:
            logger.info("Clone succeeded", repo_path=self.path)
            return CloneResult(repo)

    def _clone(self, env) -> Repo:
        try:
            return Repo.clone_from(url=self.url, to_path=self.path, branch=self.branch_name, env=env)
        except (GitError, GitCommandError) as e:
            logger.error("cannot clone policy repo: {error}", error=e)
            raise

    def _provide_git_ssh_environment(self):
        """
        provides git SSH configuration via GIT_SSH_COMMAND.

        the git ssh config will be provided only if the following conditions are met:
        - the repo url is a git ssh url
        - an ssh private key is provided in Repo Cloner __init__
        """
        if not is_ssh_repo_url(self.url) or self._ssh_key is None:
            return None # no ssh config
        git_ssh_identity_file = self._save_ssh_key_to_pem_file(self._ssh_key)
        return {
            "GIT_SSH_COMMAND": f"ssh -o StrictHostKeyChecking=no -o IdentitiesOnly=yes -i {git_ssh_identity_file}",
            "GIT_TRACE": "1",
            "GIT_CURL_VERBOSE": "1"
        }

    def _save_ssh_key_to_pem_file(self, key: str) -> Path:
        key = key.replace("_", "\n")
        if not key.endswith('\n'):
            key = key + '\n' # pem file must end with newline
        key_path = os.path.expanduser(self._ssh_key_file_path)
        parent_directory = os.path.dirname(key_path)
        if not os.path.exists(parent_directory):
            os.makedirs(parent_directory, exist_ok=True)
        with open(key_path, 'w') as f:
            f.write(key)
        os.chmod(key_path, 0o600)
        return Path(key_path)



from pathlib import Path
from sys import prefix
from .confi import Confi, confi

from opal_common.authentication.types import EncryptionKeyFormat, JWTAlgorithm

_LOG_FORMAT_WITHOUT_PID = "<green>{time}</green> | <blue>{name: <40}</blue>|<level>{level:^6} | {message}</level>\n{exception}"
_LOG_FORMAT_WITH_PID = "<green>{time}</green> | {process} | <blue>{name: <40}</blue>|<level>{level:^6} | {message}</level>\n{exception}"

class OpalCommonConfig(Confi):
    ALLOWED_ORIGINS = confi.list("ALLOWED_ORIGINS", ["*"])
    # Process name to show in logs - Not confi-controlable on purpose
    PROCESS_NAME = ""
    # Logging
    # - Log formatting
    LOG_FORMAT_INCLUDE_PID = confi.bool("LOG_FORMAT_INCLUDE_PID", False)
    LOG_FORMAT = confi.str(
        "LOG_FORMAT",
        confi.delay(
            lambda LOG_FORMAT_INCLUDE_PID=False:
            _LOG_FORMAT_WITH_PID if LOG_FORMAT_INCLUDE_PID else _LOG_FORMAT_WITHOUT_PID
        )
    )
    LOG_TRACEBACK = confi.bool("LOG_TRACEBACK", True)
    LOG_DIAGNOSE = confi.bool("LOG_DIAGNOSE", True)
    LOG_COLORIZE = confi.bool("LOG_COLORIZE", True)
    LOG_SHOW_CODE_LINE = confi.bool("LOG_SHOW_CODE_LINE", True)
    #  - log level
    LOG_LEVEL = confi.str("LOG_LEVEL", "INFO")
    #  - Which modules should be logged
    LOG_MODULE_EXCLUDE_LIST = confi.list("LOG_MODULE_EXCLUDE_LIST", [
        "uvicorn",
        # NOTE: the env var LOG_MODULE_EXCLUDE_OPA affects this list
    ])
    LOG_MODULE_INCLUDE_LIST = confi.list("LOG_MODULE_INCLUDE_LIST", ["uvicorn.protocols.http"])
    LOG_PATCH_UVICORN_LOGS = confi.bool("LOG_PATCH_UVICORN_LOGS", True,
                                        description="Should we takeover UVICORN's logs so they appear in the main logger")
    # - Log to file as well ( @see https://github.com/Delgan/loguru#easier-file-logging-with-rotation--retention--compression)
    LOG_TO_FILE = confi.bool("LOG_TO_FILE", False, description="Should we log to a file")
    LOG_FILE_PATH = confi.str("LOG_FILE_PATH", f"opal_{PROCESS_NAME}{{time}}.log", description="path to save log file")
    LOG_FILE_ROTATION = confi.str("LOG_FILE_ROTATION", "250 MB")
    LOG_FILE_RETENTION = confi.str("LOG_FILE_RETENTION", "10 days")
    LOG_FILE_COMPRESSION = confi.str("LOG_FILE_COMPRESSION", None)
    LOG_FILE_SERIALIZE = confi.str("LOG_FILE_SERIALIZE", True)
    LOG_FILE_LEVEL = confi.str("LOG_FILE_LEVEL", "INFO")

    STATISTICS_ENABLED = confi.bool("STATISTICS_ENABLED", False, description="Set if OPAL server will collect statistics about OPAL clients may cause a small performance hit")
    STATISTICS_ADD_CLIENT_CHANNEL = confi.str("STATISTICS_ADD_CLIENT_CHANNEL", "__opal_stats_add", description="The topic to update about new OPAL clients connection")
    STATISTICS_REMOVE_CLIENT_CHANNEL = confi.str("STATISTICS_REMOVE_CLIENT_CHANNEL", "__opal_stats_rm", description="The topic to update about OPAL clients disconnection")


    # Fetching Providers
    # - where to load providers from
    FETCH_PROVIDER_MODULES = confi.list("FETCH_PROVIDER_MODULES", ["opal_common.fetcher.providers"])

    # Fetching engine
    # Max number of worker tasks handling fetch events concurrently
    FETCHING_WORKER_COUNT = confi.int("FETCHING_WORKER_COUNT", 5)
    # Time in seconds to wait on the queued fetch task.
    FETCHING_CALLBACK_TIMEOUT = confi.int("FETCHING_CALLBACK_TIMEOUT", 10)
    # Time in seconds to wait for queuing a new task (if the queue is full)
    FETCHING_ENQUEUE_TIMEOUT = confi.int("FETCHING_ENQUEUE_TIMEOUT", 10)

    GIT_SSH_KEY_FILE = confi.str("GIT_SSH_KEY_FILE", str(Path.home() / ".ssh/opal_repo_ssh_key"))

    # Trust self signed certificates (Advanced Usage - only affects OPAL client) -----------------------------
    # DO NOT change these defaults unless you absolutely know what you are doing!
    # By default, OPAL client only trusts SSL certificates that are signed by a public recognized CA (certificate authority).
    # However, sometimes (mostly in on-prem setups or in dev environments) users setup their own self-signed certificates.
    # We allow OPAL client to trust these certificates, by changing the following config vars.
    CLIENT_SELF_SIGNED_CERTIFICATES_ALLOWED = confi.bool(
        "CLIENT_SELF_SIGNED_CERTIFICATES_ALLOWED", False,
        description="Whether or not OPAL Client will trust HTTPs connections protected by self signed certificates. DO NOT USE THIS IN PRODUCTION!")
    CLIENT_SSL_CONTEXT_TRUSTED_CA_FILE = confi.str(
        "CLIENT_SSL_CONTEXT_TRUSTED_CA_FILE", None,
        description="A path to your own CA public certificate file (usually a .crt or a .pem file). Certifcates signed by this issuer will be trusted by OPAL Client. DO NOT USE THIS IN PRODUCTION!")

    # security
    AUTH_PUBLIC_KEY_FORMAT = confi.enum("AUTH_PUBLIC_KEY_FORMAT", EncryptionKeyFormat, EncryptionKeyFormat.ssh)
    AUTH_PUBLIC_KEY = confi.delay(
        lambda AUTH_PUBLIC_KEY_FORMAT=None:
            confi.public_key(
                "AUTH_PUBLIC_KEY",
                default=None,
                key_format=AUTH_PUBLIC_KEY_FORMAT
            )
    )
    AUTH_JWT_ALGORITHM = confi.enum(
        "AUTH_JWT_ALGORITHM",
        JWTAlgorithm,
        getattr(JWTAlgorithm, "RS256"),
        description="jwt algorithm, possible values: see: https://pyjwt.readthedocs.io/en/stable/algorithms.html"
    )
    AUTH_JWT_AUDIENCE = confi.str("AUTH_JWT_AUDIENCE", "https://api.authorizon.com/v1/")
    AUTH_JWT_ISSUER = confi.str("AUTH_JWT_ISSUER", f"https://authorizon.com/")


opal_common_config = OpalCommonConfig(prefix="OPAL_")



import pytest
import os
import sys

# Add root opal dir to use local src as package for tests (i.e, no need for python -m pytest)
root_dir = os.path.abspath(
    os.path.join(
        os.path.dirname(__file__),
        os.path.pardir,
        os.path.pardir,
    )
)
sys.path.append(root_dir)

from pathlib import Path
from typing import List
from opal_common.urls import set_url_query_param

def test_set_url_query_param():
    base_url = 'api.authorizon.com/opal/data/config'

    # https scheme, query string not empty
    assert set_url_query_param(
        f"https://{base_url}?some=val&other=val2",
        "token",
        "secret"
    ) == f"https://{base_url}?some=val&other=val2&token=secret"

    # http scheme, query string empty
    assert set_url_query_param(f"http://{base_url}", "token", "secret") == f"http://{base_url}?token=secret"

    # no scheme, query string empty
    assert set_url_query_param(f"{base_url}", "token", "secret") == f"{base_url}?token=secret"

    # no scheme, query string not empty
    assert set_url_query_param(f"{base_url}?some=val", "token", "secret") == f"{base_url}?some=val&token=secret"




import asyncio
from typing import Callable, Coroutine, List, Optional
from git.objects import Commit

from opal_common.git.branch_tracker import BranchTracker
from opal_common.git.repo_cloner import RepoCloner
from opal_common.git.exceptions import GitFailed
from opal_common.logger import logger
from opal_common.sources.base_policy_source import BasePolicySource


class GitPolicySource(BasePolicySource):
    """
    Watches a git repository for changes and can trigger callbacks
    when detecting new commits on the tracked branch.

    Checking for changes is done following a git pull from a tracked
    remote. The pull can be either triggered by a method (i.e: you can
    call it from a webhook) or can be triggered periodically by a polling
    task.

    Args:
        remote_source_url(str): the base address to request the policy from
        local_clone_path(str):  path for the local git to manage policies
        branch_name(str):  name of remote branch in git to pull, default to master
        ssh_key (str, optional): private ssh key used to gain access to the cloned repo
        polling_interval(int):  how many seconds need to wait between polling
        request_timeout(int):  how many seconds need to wait until timout

    """
    def __init__(
        self,
        remote_source_url: str,
        local_clone_path: str,
        branch_name: str = "master",
        ssh_key: Optional[str] = None,
        polling_interval: int = 0,
        request_timeout: int = 0,
    ):
        super().__init__(remote_source_url=remote_source_url, local_clone_path=local_clone_path,
                         polling_interval=polling_interval)

        self._cloner = RepoCloner(remote_source_url, local_clone_path, branch_name=branch_name, ssh_key=ssh_key, clone_timeout=request_timeout)
        self._branch_name = branch_name
        self._tracker = None

    async def get_initial_policy_state_from_remote(self):
        """
        init remote data to local repo
        """
        try:
            result = await self._cloner.clone()
        except GitFailed as e:
            await self._on_git_failed(e)
            return

        self._tracker = BranchTracker(
            repo=result.repo,
            branch_name=self._branch_name,
        )

    async def check_for_changes(self):
        """
        Calling this method will trigger a git pull from the tracked remote.
        If after the pull the watcher detects new commits, it will call the
        callbacks registered with _on_new_policy().
        """
        logger.info("Pulling changes from remote: '{remote}'", remote=self._tracker.tracked_remote.name)
        has_changes, prev, latest = self._tracker.pull()
        if not has_changes:
            logger.info("No new commits: HEAD is at '{head}'", head=latest.hexsha)
        else:
            logger.info("Found new commits: old HEAD was '{prev_head}', new HEAD is '{new_head}'",
                        prev_head=prev.hexsha, new_head=latest.hexsha)
            await self._on_new_policy(old=prev, new=latest)



class Formatter:

    MAX_FIELD_LEN = 25

    def __init__(self, format_string: str):
        self.fmt = format_string

    def limit_len(self, record, field, length=MAX_FIELD_LEN):
        #Shorten field content
        content = record[field]
        if len(content) > length:
            parts = content.split(".")
            if len(parts) > 2:
                content = f"{parts[0]}...{parts[-1]}"
        if len(content) > length:
            content = f"{content[:length-3]}..."
        record[field] = content

    def format(self, record):
        self.limit_len(record,"name",40)
        return self.fmt




from typing import List, Set, Union
from pathlib import Path

from opal_common.utils import sorted_list_from_set

class PathUtils:
    @staticmethod
    def intermediate_directories(paths: List[Path]) -> List[Path]:
        """
        returns the set of all parent directories for a list of paths.
        i.e: calculate all partial paths that are directories.
        """
        directories = set()
        for path in paths:
            directories.update(path.parents)
        return sorted_list_from_set(directories)

    @staticmethod
    def is_child_of_directories(path: Path, directories: Set[Path]) -> bool:
        """
        whether the input path is a child of one of the input directories
        """
        return bool(directories & set(path.parents))

    @staticmethod
    def filter_children_paths_of_directories(paths: List[Path], directories: Set[Path]) -> List[Path]:
        """
        returns only paths in :paths that are children of one of the paths in :directories.
        """
        return [path for path in paths if PathUtils.is_child_of_directories(path, directories)]

    @staticmethod
    def non_intersecting_directories(paths: List[Path]) -> Set[Path]:
        """
        gets a list of paths (directories), and returns a set of directories that are non-intersecting,
        meaning no directory in the set is a parent of another directory in the set (i.e: parent
        directories "swallow" their subdirectories).
        """
        output_paths = set()
        for candidate in paths:
            if set(candidate.parents) & output_paths:
                # the next candidate is covered by a parent which is already in output -> SKIP
                # or the next candidate is already in the list
                continue
            for out_path in list(output_paths):
                # the next candidate can displace a child from the output
                if candidate in list(out_path.parents):
                    output_paths.remove(out_path)
            output_paths.add(candidate)
        return output_paths

    @staticmethod
    def sort_paths_according_to_explicit_sorting(unsorted_paths: List[Path], explicit_sorting: List[Path]) -> List[Path]:
        """
        the way this sorting works, is assuming that explicit_sorting does NOT necessarily contains all the paths
        found in the original list. We must ensure that all items in unsorted_paths must also exist in the output list.
        """
        unsorted = unsorted_paths.copy()

        sorted_paths: List[Path] = []
        for path in explicit_sorting:
            try:
                # we look for Path objects and not str for normalization of the path
                found_path: Path = unsorted.pop(unsorted.index(path))
                sorted_paths.append(found_path)
            except ValueError:
                continue # skip, not found in the original list

        # add the remaineder to the end of the sorted list
        sorted_paths.extend(unsorted)

        return sorted_paths


import logging
import os
import sys

# Add parent path to use local src as package for tests
root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir, os.path.pardir, os.path.pardir ))
sys.path.append(root_dir)

import asyncio
from multiprocessing import Process

import pytest
import uvicorn
from fastapi import FastAPI, Depends, Header, HTTPException

from opal_common.fetcher import FetchingEngine
from opal_common.fetcher.providers.http_fetch_provider import HttpFetcherConfig


# Configurable
PORT = int(os.environ.get("PORT") or "9110")
BASE_URL = f"http://localhost:{PORT}"
DATA_ROUTE = f"/data"
AUTHORIZED_DATA_ROUTE = f"/data_authz"
SECRET_TOKEN = "fake-super-secret-token"
DATA_KEY = "Hello"
DATA_VALUE = "World"
DATA_SECRET_VALUE = "SecretWorld"

async def check_token_header(x_token: str = Header(None)):
    if x_token != SECRET_TOKEN:
        raise HTTPException(status_code=400, detail="X-Token header invalid")
    return None

def setup_server():
    app =  FastAPI()

    @app.get(DATA_ROUTE)
    def get_data():
        return {DATA_KEY: DATA_VALUE}

    @app.get(AUTHORIZED_DATA_ROUTE)
    def get_authorized_data(token=Depends(check_token_header)):
        return {DATA_KEY: DATA_SECRET_VALUE}

    uvicorn.run(app, port=PORT )

@pytest.fixture(scope="module")
def server():
    # Run the server as a separate process
    proc = Process(target=setup_server, args=(), daemon=True)
    proc.start()
    yield proc
    proc.kill() # Cleanup after test

@pytest.mark.asyncio
async def test_simple_http_get(server):
    """
    Simple http get
    """
    got_data_event = asyncio.Event()
    async with FetchingEngine() as engine:
        async def callback(data):
            assert data[DATA_KEY] == DATA_VALUE
            got_data_event.set()
        await engine.queue_url(f"{BASE_URL}{DATA_ROUTE}", callback)
        await asyncio.wait_for(got_data_event.wait(), 5)
        assert got_data_event.is_set()

@pytest.mark.asyncio
async def test_simple_http_get_with_wait(server):
    """
    Simple http get - with 'queue_url_and_wait'
    """
    async with FetchingEngine() as engine:
        data = await engine.handle_url(f"{BASE_URL}{DATA_ROUTE}")
        assert data[DATA_KEY] == DATA_VALUE


@pytest.mark.asyncio
async def test_authorized_http_get(server):
    """
    Test getting data from a server route with an auth token
    """
    got_data_event = asyncio.Event()
    async with FetchingEngine() as engine:
        async def callback(data):
            assert data[DATA_KEY] == DATA_SECRET_VALUE
            got_data_event.set()
        # fetch with bearer token authorization
        await engine.queue_url(f"{BASE_URL}{AUTHORIZED_DATA_ROUTE}", callback, HttpFetcherConfig(headers={"X-TOKEN": SECRET_TOKEN}))
        await asyncio.wait_for(got_data_event.wait(), 5)
        assert got_data_event.is_set()

@pytest.mark.asyncio
async def test_authorized_http_get_from_dict(server):
    """
    Just like test_authorized_http_get, but we also check that the FetcherConfig is adapted from "the wire" (as a dict instead of the explicit HttpFetcherConfig)
    """
    got_data_event = asyncio.Event()
    async with FetchingEngine() as engine:
        async def callback(data):
            assert data[DATA_KEY] == DATA_SECRET_VALUE
            got_data_event.set()
        # raw config to be parsed
        config = {
            "headers" : {"X-TOKEN": SECRET_TOKEN}
        }
        # fetch with bearer token authorization
        await engine.queue_url(f"{BASE_URL}{AUTHORIZED_DATA_ROUTE}", callback, config)
        await asyncio.wait_for(got_data_event.wait(), 5)
        assert got_data_event.is_set()


@pytest.mark.asyncio
async def test_external_http_get():
    """
    Test simple http get on external (https://freegeoip.app/) site
    Checking we get a JSON with the data we expected (the IP we queried)
    """
    got_data_event = asyncio.Event()
    async with FetchingEngine() as engine:
        async def callback(data):
            assert data["ip"] == "8.8.8.8"
            got_data_event.set()
        await engine.queue_url(f"https://freegeoip.app/json/8.8.8.8", callback)
        await asyncio.wait_for(got_data_event.wait(), 5)
        assert got_data_event.is_set()



import pytest
import os
import sys

# Add root opal dir to use local src as package for tests (i.e, no need for python -m pytest)
root_dir = os.path.abspath(
    os.path.join(
        os.path.dirname(__file__),
        os.path.pardir,
        os.path.pardir,
    )
)
sys.path.append(root_dir)

from pathlib import Path
from typing import List
from opal_common.paths import PathUtils


def to_paths(paths: List[str]) -> List[Path]:
    return [Path(path) for path in paths]

def test_intermediate_directories():
    # empty sources returns empty parent list
    assert len(PathUtils.intermediate_directories(to_paths([]))) == 0
    # '/', '.' and '' has no parent
    assert len(PathUtils.intermediate_directories(to_paths(['/']))) == 0
    assert len(PathUtils.intermediate_directories(to_paths(['.']))) == 0
    assert len(PathUtils.intermediate_directories(to_paths(['']))) == 0
    # top level directories has only one parent
    assert PathUtils.intermediate_directories(to_paths(['/some'])) == to_paths(['/'])
    assert PathUtils.intermediate_directories(to_paths(['some'])) == to_paths(['.'])
    # check some examples of nested paths
    parents = PathUtils.intermediate_directories(to_paths(['some/dir/to']))
    assert len(parents) == 3
    assert len(set(parents).intersection(set(to_paths(['.', 'some', 'some/dir'])))) == 3
    parents = PathUtils.intermediate_directories(to_paths(['/another/example']))
    assert len(parents) == 2
    assert len(set(parents).intersection(set(to_paths(['/', '/another'])))) == 2
    # mix and match
    parents = PathUtils.intermediate_directories(to_paths([
        'some',
        '/other',
        'example/of/path',
        'some/may/intersect',
    ]))
    assert len(parents) == 6
    assert Path('.') in parents
    assert Path('/') in parents
    assert Path('some') in parents
    assert Path('some/may') in parents
    assert Path('example') in parents
    assert Path('example/of') in parents


def test_is_child_of_directories():
    # parent directories are the top level (relative) dir
    assert PathUtils.is_child_of_directories(Path('.'), set(to_paths(['.']))) == False
    assert PathUtils.is_child_of_directories(Path('hello'), set(to_paths(['.']))) == True
    assert PathUtils.is_child_of_directories(Path('world.txt'), set(to_paths(['.']))) == True
    assert PathUtils.is_child_of_directories(Path('/world'), set(to_paths(['.']))) == False

    # parent directories are the top level (absolute) dir
    assert PathUtils.is_child_of_directories(Path('/'), set(to_paths(['/']))) == False
    assert PathUtils.is_child_of_directories(Path('/hello'), set(to_paths(['/']))) == True
    assert PathUtils.is_child_of_directories(Path('/world.txt'), set(to_paths(['/']))) == True
    assert PathUtils.is_child_of_directories(Path('world'), set(to_paths(['/']))) == False

    # directories can be files (bad input)
    assert PathUtils.is_child_of_directories(Path('/world.txt'), set(to_paths(['/hello.txt']))) == False

    # some valid input
    assert PathUtils.is_child_of_directories(Path('some/file.txt'), set(to_paths(['some']))) == True
    assert PathUtils.is_child_of_directories(Path('some/file.txt'), set(to_paths(['.']))) == True
    assert PathUtils.is_child_of_directories(Path('some/dir/to/file.txt'), set(to_paths(['some/dir']))) == True

def test_filter_children_paths_of_directories():
    sources = to_paths([
        '/files/for/testing/1.txt',
        '/files/for/testing/2.json',
        '/filtered/out.txt',
        'relative/path.log',
        'relative/subdir/another.log',
    ])
    # filter paths under .
    paths = PathUtils.filter_children_paths_of_directories(sources, set(to_paths(['.'])))
    assert len(paths) == 2
    assert len(set(paths).intersection(set(to_paths([
        'relative/path.log',
        'relative/subdir/another.log'
    ])))) == 2

    # filter paths under /
    paths = PathUtils.filter_children_paths_of_directories(sources, set(to_paths(['/'])))
    assert len(paths) == 3
    assert len(set(paths).intersection(set(to_paths([
        '/files/for/testing/1.txt',
        '/files/for/testing/2.json',
        '/filtered/out.txt'
    ])))) == 3

    # filter paths under /files
    paths = PathUtils.filter_children_paths_of_directories(sources, set(to_paths(['/files'])))
    assert len(paths) == 2
    assert len(set(paths).intersection(set(to_paths([
        '/files/for/testing/1.txt',
        '/files/for/testing/2.json',
    ])))) == 2

    # filter paths under relative/subdir
    paths = PathUtils.filter_children_paths_of_directories(sources, set(to_paths(['relative/subdir'])))
    assert len(paths) == 1
    assert len(set(paths).intersection(set(to_paths([
        'relative/subdir/another.log',
    ])))) == 1

    # filter paths under multiple parents
    paths = PathUtils.filter_children_paths_of_directories(sources, set(to_paths(['relative/subdir', '/filtered'])))
    assert len(paths) == 2
    assert len(set(paths).intersection(set(to_paths([
        '/filtered/out.txt',
        'relative/subdir/another.log',
    ])))) == 2

    # parents can intersect
    paths = PathUtils.filter_children_paths_of_directories(sources, set(to_paths(['relative/subdir', '.'])))
    assert len(paths) == 2
    assert len(set(paths).intersection(set(to_paths([
        'relative/path.log',
        'relative/subdir/another.log'
    ])))) == 2

    # no parents
    paths = PathUtils.filter_children_paths_of_directories(sources, set())
    assert len(paths) == 0

    # no parent match sources
    paths = PathUtils.filter_children_paths_of_directories(sources, set(to_paths(['not/in/repo'])))
    assert len(paths) == 0

def test_non_intersecting_directories():
    # assert PathUtils.is_child_of_directories(Path('/hello'), set(to_paths(['/']))) == True

    # relative paths are all displaced by '.'
    assert PathUtils.non_intersecting_directories(to_paths(['.', 'hello', 'world'])) == set(to_paths(['.']))

    # absolute paths are all displaced by '/'
    assert PathUtils.non_intersecting_directories(to_paths(['/', '/hello', '/hello/world'])) == set(to_paths(['/']))

    # parents displace children
    assert PathUtils.non_intersecting_directories(
        to_paths(['/hello', '/hello/world', 'world', 'world/of/tomorrow'])) == set(to_paths(['/hello', 'world']))

def test_sort_paths_according_to_explicit_sorting():
    sort_func = PathUtils.sort_paths_according_to_explicit_sorting

    # empty list cannot be sorted
    assert sort_func([], []) == []
    assert sort_func([], to_paths(['.', 'hello', 'world'])) == []

    # if no sorting is available, returns the original list
    examples = [
        to_paths(['.', 'hello', 'world']),
        to_paths(['world.rego', 'lib.rego', './unrelated.rego']),
        to_paths(['/', './unrelated.rego']),
    ]
    for example in examples:
        assert sort_func(example, []) == example

    # partial sorting works as expected
    assert sort_func(
        to_paths(['.', 'world.rego', 'lib.rego', 'more/path.rego', 'even/more']),
        # ordering
        to_paths(['lib.rego', 'world.rego']),
    # explicitly sorted items move to the beginning of the list
    # other items remain in the original sorting
    ) == to_paths(['lib.rego', 'world.rego', '.', 'more/path.rego', 'even/more'])

    # same list, switch the ordering
    assert sort_func(
        to_paths(['.', 'world.rego', 'lib.rego', 'more/path.rego', 'even/more']),
        # ordering
        to_paths(['world.rego', 'lib.rego']),
    # explicitly sorted items move to the beginning of the list
    # other items remain in the original sorting
    ) == to_paths(['world.rego', 'lib.rego', '.', 'more/path.rego', 'even/more'])

    # some sorting items are not found
    assert sort_func(
        to_paths(['.', 'world.rego', 'lib.rego', 'more/path.rego', 'even/more']),
        # ordering
        to_paths(['world.rego', 'lib2.rego']),
    # explicitly sorted items move to the beginning of the list
    # other items remain in the original sorting
    ) == to_paths(['world.rego', '.', 'lib.rego', 'more/path.rego', 'even/more'])


import aiohttp
import os
import sys

# Add parent path to use local src as package for tests
root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir, os.path.pardir, os.path.pardir ))
print(root_dir)
sys.path.append(root_dir)

import asyncio
import tenacity
from multiprocessing import Process

import pytest
import uvicorn
from fastapi import FastAPI

from opal_common.fetcher.providers.http_fetch_provider import HttpFetchEvent, HttpFetchProvider
from opal_common.fetcher import FetchingEngine, FetchEvent

# Configurable
PORT = int(os.environ.get("PORT") or "9110")
BASE_URL = f"http://localhost:{PORT}"
DATA_ROUTE = f"/data"
DATA_KEY = "Hello"
DATA_VALUE = "World"
DATA_SECRET_VALUE = "SecretWorld"

@pytest.mark.asyncio
async def test_retry_failure():
    """
    Test callback on failure
    """
    got_data_event = asyncio.Event()
    got_error = asyncio.Event()

    async with FetchingEngine() as engine:
        # callback to handle failure
        async def error_callback(error:Exception, event:FetchEvent):
            # check we got the expection we expected
            assert isinstance(error, aiohttp.client_exceptions.ClientConnectorError)
            got_error.set()
        # register the callback
        engine.register_failure_handler(error_callback)
        # callback for success - shouldn't eb called in this test
        async def callback(result):
            got_data_event.set()
        # Use an event on an invalid port - and only to attempts
        retry_config = HttpFetchProvider.DEFAULT_RETRY_CONFIG.copy()
        retry_config["stop"] = tenacity.stop.stop_after_attempt(2)
        event = HttpFetchEvent(url=f"http://localhost:25", retry=retry_config)
        # queue the event
        await engine.queue_fetch_event(event, callback)
        # wait for the failure callback
        await asyncio.wait_for(got_error.wait(), 25)
        assert not got_data_event.is_set()
        assert got_error.is_set()



import asyncio
import uuid
from typing import Coroutine, Dict, List, Union

from ..events import FetcherConfig, FetchEvent
from ..fetch_provider import BaseFetchProvider
from ..fetcher_register import FetcherRegister
from ..logger import get_logger
from .base_fetching_engine import BaseFetchingEngine
from .fetch_worker import fetch_worker
from .core_callbacks import OnFetchFailureCallback


logger = get_logger("engine")


class FetchingEngine(BaseFetchingEngine):
    """
    A Task queue manager for fetching events.

    - Configure with different fetcher providers - via __init__'s register_config or via self.register.register_fetcher()
    - Use queue_url() to fetch a given URL with the default FetchProvider
    - Use queue_fetch_event() to fetch data using a configured FetchProvider
    - Use with 'async with' to terminate tasks (or call self.terminate_tasks() when done)
    """

    DEFAULT_WORKER_COUNT = 5
    DEFAULT_CALLBACK_TIMEOUT = 10
    DEFAULT_ENQUEUE_TIMEOUT = 10

    @staticmethod
    def gen_uid():
        return uuid.uuid4().hex

    def __init__(self, register_config: Dict[str, BaseFetchProvider] = None,
                 worker_count: int = DEFAULT_WORKER_COUNT,
                 callback_timeout: int = DEFAULT_CALLBACK_TIMEOUT,
                 enqueue_timeout: int = DEFAULT_ENQUEUE_TIMEOUT
                 ) -> None:
        # The internal task queue (created at start_workers)
        self._queue: asyncio.Queue = None
        # Worker working the queue
        self._tasks: List[asyncio.Task] = []
        # register of the fetch providers workers can use
        self._fetcher_register = FetcherRegister(register_config)
        # core event callback regsiters
        self._failure_handlers: List[OnFetchFailureCallback] = []
        # how many workers to run
        self._worker_count: int = worker_count
        # time in seconds before timeout on a fetch callback
        self._callback_timeout = callback_timeout
        # time in seconds before time out on adding a task to queue (when full)
        self._enqueue_timeout = enqueue_timeout

    def start_workers(self):
        if self._queue is None:
            self._queue = asyncio.Queue()
            # create worker tasks
            for _ in range(self._worker_count):
                self.create_worker()

    @property
    def register(self) -> FetcherRegister:
        return self._fetcher_register

    async def __aenter__(self):
        """
        Async Context manager to cancel tasks on exit 
        """
        self.start_workers()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        if (exc is not None):
            logger.error("Error occurred within FetchingEngine context", exc_info=repr((exc_type, exc, tb)))
        await self.terminate_workers()

    async def terminate_workers(self):
        """
        Cancel and wait on the internal worker tasks
        """
        # Cancel our worker tasks.
        for task in self._tasks:
            task.cancel()
        # Wait until all worker tasks are cancelled.
        await asyncio.gather(*self._tasks, return_exceptions=True)
        # reset queue
        self._queue = None

    async def handle_url(self, url: str, timeout: float = None, **kwargs):
        """
        Same as self.queue_url but instead of using a callback, you can wait on this coroutine for the result as a return value
        Args:
            url (str): 
            timeout (float, optional): time in seconds to wait on the queued fetch task. Defaults to self._callback_timeout.
            kwargs: additional args passed to self.queue_url

        Raises:
            asyncio.TimeoutError: if the given timeout has expired
            also - @see self.queue_fetch_event            
        """
        timeout = self._callback_timeout if timeout is None else timeout
        wait_event = asyncio.Event()
        data = {'result': None}
        # Callback to wait and retrive data

        async def waiter_callback(answer):
            data['result'] = answer
            # Signal callback is done
            wait_event.set()
        await self.queue_url(url, waiter_callback, **kwargs)
        # Wait with timeout
        if timeout is not None:
            await asyncio.wait_for(wait_event.wait(), timeout)
        # wait forever
        else:
            await wait_event.wait()
        # return saved result value from callback
        return data['result']

    async def queue_url(self, url: str, callback: Coroutine, config: Union[FetcherConfig, dict] = None, fetcher="HttpFetchProvider") -> FetchEvent:
        """
        Simplified default fetching handler for queuing a fetch task

        Args:
            url (str): the URL to fetch from
            callback (Coroutine): a callback to call with the fetched result
            config (FetcherConfig, optional): Configuration to be used by the fetcher. Defaults to None.
            fetcher (str, optional): Which fetcher class to use. Defaults to "HttpFetchProvider".
        Returns: 
            the queued event (which will be mutated to at least have an Id)

        Raises:
            @see self.queue_fetch_event
        """
        # override default fetcher with (potential) override value from FetcherConfig
        if isinstance(config, dict) and config.get("fetcher", None) is not None:
            fetcher = config["fetcher"]
        elif isinstance(config, FetcherConfig) and config.fetcher is not None:
            fetcher = config.fetcher

        # init a URL event
        event = FetchEvent(url=url, fetcher=fetcher, config=config)
        return await self.queue_fetch_event(event, callback)

    async def queue_fetch_event(self, event: FetchEvent, callback: Coroutine, enqueue_timeout=None) -> FetchEvent:
        """
        Basic handler to queue a fetch event for a fetcher class.
        Waits if the queue is full until enqueue_timeout seconds; if enqueue_timeout is None returns immediately or raises QueueFull

        Args:
            event (FetchEvent): the fetch event to queue as a task
            callback (Coroutine): a callback to call with the fetched result
            enqueue_timeout (float): timeout in seconds or None for no timeout, Defaults to self.DEFAULT_ENQUEUE_TIMEOUT

        Returns: 
            the queued event (which will be mutated to at least have an Id)  

        Raises:
            asyncio.QueueFull: if the queue is full and enqueue_timeout is set as None
            asyncio.TimeoutError: if enqueue_timeout is not None, and the queue is full and hasn't cleared by the timeout time
        """
        enqueue_timeout = enqueue_timeout if enqueue_timeout is not None else self._enqueue_timeout
        # Assign a unique identifier for the event
        event.id = self.gen_uid()
        # add to the queue for handling
        # if no timeout we return immediately or raise QueueFull
        if enqueue_timeout is None:
            await self._queue.put_nowait((event, callback))
        # if timeout
        else:
            await asyncio.wait_for(self._queue.put((event, callback)), enqueue_timeout)
        return event

    def create_worker(self) -> asyncio.Task:
        """
        Create an asyncio worker to work the engine's queue
        Engine init starts several workers according to given configuration
        """
        task = asyncio.create_task(fetch_worker(self._queue, self))
        self._tasks.append(task)
        return task

    def register_failure_handler(self, callback: OnFetchFailureCallback):
        """
        Register a callback to be called with exception and original event in case of failure

        Args:
            callback (OnFetchFailureCallback): callback to register
        """
        self._failure_handlers.append(callback)

    async def _management_event_handler(self, handlers: List[Coroutine], *args, **kwargs):
        """
        Generic management event subscriber caller
        Args:
            handlers (List[Coroutine]): callback coroutines
        """
        await asyncio.gather(*(callback(*args, **kwargs) for callback in handlers))

    async def _on_failure(self, error: Exception, event: FetchEvent):
        """
        Call event failure subscribers

        Args:
            error (Exception): thrown exception
            event (FetchEvent): event which was being handled
        """
        await self._management_event_handler(self._failure_handlers, error, event)



import logging

def hijack_uvicorn_logs(intercept_handler: logging.Handler):
    """
    Uvicorn loggers are configured to use special handlers.

    Adding an intercept handler to the root logger manages to intercept logs from uvicorn, however, the log messages are duplicated.
    This is happening because uvicorn loggers are propagated by default - we get a log message once for the "uvicorn" / "uvicorn.error"
    logger and once for the root logger). Another stupid issue is that the "uvicorn.error" logger is not just for errors, which is confusing.

    This method is doing 2 things for each uvicorn logger:
    1) remove all existing handlers and replace them with the intercept handler (i.e: will be logged via loguru)
    2) cancel propagation - which will mean messages will not propagate to the root logger (which also has an InterceptHandler), fixing the duplication
    """
    # get loggers directly from uvicorn config - if they will change something - we will know.
    from uvicorn.config import LOGGING_CONFIG
    uvicorn_logger_names = list(LOGGING_CONFIG.get("loggers", {}).keys()) or ["uvicorn", "uvicorn.access", "uvicorn.error"]
    for logger_name in uvicorn_logger_names:
        logger = logging.getLogger(logger_name)
        logger.handlers = [intercept_handler]
        logger.propagate = False



from typing import Optional
from uuid import UUID

from fastapi import Header
from fastapi.exceptions import HTTPException
from fastapi.security.utils import get_authorization_scheme_param

from opal_common.authentication.types import JWTClaims
from opal_common.authentication.verifier import JWTVerifier, Unauthorized
from opal_common.logger import logger

def get_token_from_header(authorization_header: str) -> Optional[str]:
    """
    extracts a bearer token from an HTTP Authorization header.

    when provided bearer token via websocket,
    we cannot use the fastapi built-in: oauth2_scheme.
    """
    if not authorization_header:
        return None

    scheme, token = get_authorization_scheme_param(authorization_header)
    if not token or scheme.lower() != "bearer":
        return None

    return token


def verify_logged_in(verifier: JWTVerifier, token: Optional[str]) -> JWTClaims:
    """
    forces bearer token authentication with valid JWT or throws 401.
    """
    try:
        if not verifier.enabled:
            logger.debug("JWT verification disabled, cannot verify requests!")
            return {}
        if token is None:
            raise Unauthorized(token=token, description="access token was not provided")
        claims: JWTClaims = verifier.verify(token)
        subject = claims.get("sub", "")

        invalid = Unauthorized(token=token, description="invalid sub claim")
        if not subject:
            raise invalid
        try:
            _ = UUID(subject)
        except ValueError:
            raise invalid

        # returns the entire claims dict so we can do more checks on it if needed
        return claims or {}

    except (Unauthorized, HTTPException) as err:
        # err.details is sometimes string and sometimes dict
        details: dict = {}
        if isinstance(err.detail, dict):
            details = err.detail.copy()
        elif isinstance(err.detail, str):
            details = {"msg": err.detail}
        else:
            details = {"msg": repr(err.detail)}

        # pop the token before logging - tokens should not appear in logs
        details.pop("token", None)

        # logs the error and reraises
        logger.error(f"Authentication failed with {err.status_code} due to error: {details}")
        raise


class _JWTAuthenticator:
    def __init__(self, verifier: JWTVerifier):
        self._verifier = verifier

    @property
    def verifier(self) -> JWTVerifier:
        return self._verifier

    @property
    def enabled(self) -> JWTVerifier:
        return self._verifier.enabled

class JWTAuthenticator(_JWTAuthenticator):
    """
    bearer token authentication for http(s) api endpoints.
    throws 401 if a valid jwt is not provided.
    """
    def __call__(self, authorization: Optional[str] = Header(None)) -> JWTClaims:
        token = get_token_from_header(authorization)
        return verify_logged_in(self._verifier, token)


class WebsocketJWTAuthenticator(_JWTAuthenticator):
    """
    bearer token authentication for websocket endpoints.

    with fastapi ws endpoint, we cannot throw http exceptions inside dependencies,
    because no matter the http status code, uvicorn will treat it as http 500.
    see: https://github.com/encode/uvicorn/blob/master/uvicorn/protocols/websockets/websockets_impl.py#L168

    Instead we return a boolean to the endpoint, in order for it to gracefully
    close the connection in case authentication was unsuccessful.

    In this case uvicorn's hardcoded behavior suits us:
    - if websocket.accept() was called, http 200 will be sent
    - if websocket.close() was called instead, http 403 will be sent
    no other status code are supported.
    see: https://github.com/encode/uvicorn/blob/master/uvicorn/protocols/websockets/websockets_impl.py#L189-L207

    thus we return a boolean and the endpoint can use it to potentially call websocket.close()
    """
    def __call__(self, authorization: Optional[str] = Header(None)) -> bool:
        token = get_token_from_header(authorization)
        try:
            verify_logged_in(self._verifier, token)
            return True
        except (Unauthorized, HTTPException):
            return False


class StaticBearerAuthenticator:
    """
    bearer token authentication for http(s) api endpoints.
    throws 401 if token does not match a preconfigured value.
    """
    def __init__(self, preconfigured_token: Optional[str]):
        self._preconfigured_token = preconfigured_token

    def __call__(self, authorization: Optional[str] = Header(None)):
        if self._preconfigured_token is None:
            # always allow
            return

        if authorization is None:
            raise Unauthorized(description="Authorization header is required!")

        token = get_token_from_header(authorization)
        if token is None or token != self._preconfigured_token:
            raise Unauthorized(token=token, description="unauthorized to access this endpoint!")



from opal_common.schemas.security import PeerType
from opal_common.authentication.deps import JWTAuthenticator
from opal_common.authentication.verifier import Unauthorized
from opal_common.authentication.types import JWTClaims

def require_peer_type(authenticator: JWTAuthenticator, claims: JWTClaims, required_type: PeerType):
    if not authenticator.enabled:
        return

    peer_type = claims.get('peer_type', None)
    if peer_type is None:
        raise Unauthorized(description="Missing 'peer_type' claim for OPAL jwt token")
    try:
        type = PeerType(peer_type)
    except ValueError:
        raise Unauthorized(description=f"Invalid 'peer_type' claim for OPAL jwt token: {peer_type}")

    if type != required_type:
        raise Unauthorized(description=f"Incorrect 'peer_type' claim for OPAL jwt token: {str(type)}, expected: {str(required_type)}")


"""
Easy Python configuration interface built on top of python-decouple and click / typer.
Adding typing support and parsing with Pydantic and Enum.
"""

import inspect
from collections import OrderedDict
from typing import Callable, Dict, List, Tuple, TypeVar, Optional, Any, Union
from functools import partial, wraps
from pydantic import BaseModel
from decouple import config, Csv, text_type, undefined, UndefinedValueError
import string
import json

from typer import Typer

from .types import ConfiDelay, ConfiEntry, no_cast
from .cli import get_cli_object_for_config_objects


from opal_common.authentication.types import EncryptionKeyFormat, PrivateKey, PublicKey
from opal_common.authentication.casting import cast_private_key, cast_public_key

def cast_boolean(value):
    """
    Parse an entry as a boolean.
     - all variations of "true" and 1 are treated as True
     - all variations of "false" and 0 are treated as False
    """
    if isinstance(value, bool):
        return value
    elif isinstance(value, str):
        value = value.lower()
        if (value == "true" or value == "1"):
            return True
        elif value == "false" or value == "0":
            return False
        else:
            raise UndefinedValueError(f"{value} - is not a valid boolean")
    else:
        raise UndefinedValueError(f"{value} - is not a valid boolean")

def cast_pydantic(model:BaseModel):
    def cast_pydantic_by_model(value):
        if isinstance(value, str):
            return model.parse_raw(value)
        else:
            return model.parse_obj(value)
    return cast_pydantic_by_model

def ignore_confi_delay_cast(cast_func):
    """
    when we pass a ConfiDelay as the default to decouple, until
    this delayed default is evaluated by confi, there is no point in
    casting it.

    After a ConfiDelay is evaluated, the resulted value should be passed
    again to the cast method, and this time it will indeed be cast.
    """
    @wraps(cast_func)
    def wrapped_cast(value, *args, **kwargs):
        if isinstance(value, ConfiDelay):
            return value
        return cast_func(value, *args, **kwargs)
    return wrapped_cast


def load_conf_if_none(variable, conf):
    if variable is None:
        return conf
    else:
        return variable


EnumT = TypeVar("EnumT")
T = TypeVar("T", bound=BaseModel)
ValueT = TypeVar("ValueT")


class Confi:
    """
    Interface to create typed configuration entries
    """

    def __init__(self, prefix=None, is_model=True) -> None:
        """

        Args:
            prefix (str, optional): Prefix to add to all env-var keys. Defaults to self.ENV_PREFIX (which defaults to "").
            is_model (bool, optional): Should Confi.<type> return a ConfiEntry (the default, True) or should it evaluate env settings immediately and return a value (False)
        """
        self._is_model = is_model
        self._prefix = prefix
        # counter of created entries (to track order)
        self._counter = 0
        # entries to be evaluated
        self._entries: Dict[str, ConfiEntry] = OrderedDict()
        # delayed entries to be evaluated (instead of being referenced by self._entries)
        self._delayed_entries: Dict[str, ConfiDelay] = OrderedDict()
        # entries with delayed defaults (in addition to being referenced by self._entries)
        self._delayed_defaults: Dict[str, ConfiEntry] = OrderedDict()

        # get members by creation order
        members = sorted(inspect.getmembers(self, self._is_entry), key=self._get_entry_index)
        # eval class entries into values (by order of defintion - same order as in the config class lines)
        for name, entry in members:
            # unwrap delayed entries
            if isinstance(entry, ConfiDelay):
                entry = entry.eval(self)

            if isinstance(entry, ConfiEntry):
                self._entries[name] = entry
                # save delayed
                if isinstance(entry.default, ConfiDelay):
                    self._delayed_defaults[name] = entry
                # eval, and save the value into the class instance
                value = self._eval_and_save_entry(name, entry)
                # save the value into the entry to be used as default for CLI
                entry.value = value

        # load (all calls inside should produce a real value)
        self._is_model = False

        # load delayed values:
        for name, entry in self._delayed_defaults.items():
            default: ConfiDelay =  entry.default
            # but only if no value is set yet
            if entry.value == default or entry.value == undefined:
                setattr(self, name, entry.cast(default.eval(self)))

        self.on_load()
        self._is_model = is_model

    def _is_entry(self, entry):
        res = isinstance(entry, (ConfiEntry,ConfiDelay))
        return res

    def _get_entry_index(self, member:Tuple[str,ConfiEntry]):
        name, entry = member
        return entry.index


    @property
    def entries(self):
        return self._entries

    def _prefix_key(self, key):
        prefix = self._prefix
        return f"{prefix}{key}" if prefix is not None else key

    def _eval_and_save_entry(self, name:str, entry:ConfiEntry):
        value = self._eval_entry(entry)
        setattr(self, name, value)
        return value

    def _eval_entry(self, entry:ConfiEntry):
        whole_key = self._prefix_key(entry.key)
        res = self._evaluate(whole_key,entry.default,entry.cast,**entry.kwargs)
        return res

    def _process(self, key, default=undefined, description=None, cast=no_cast, type:ValueT=str, flags:List[str]=None, **kwargs) -> Union[ValueT, ConfiEntry]:
        if self._is_model:
            # create new entry
            res = ConfiEntry(key, default,description, cast, type, index=self._counter, flags=flags, **kwargs)
            # track count for indexing
            self._counter +=1
            return res
        else:
            whole_key = self._prefix_key(key)
            return self._evaluate(whole_key, default, cast, **kwargs)

    def _evaluate(self, key, default=undefined, cast=no_cast, **kwargs):
        safe_cast_func = ignore_confi_delay_cast(cast)
        try:
            res = config(key, default=default, cast=safe_cast_func, **kwargs)
        except:
            if default is undefined:
                raise
            return default
        return res

    def __repr__(self) -> str:
        return json.dumps({k:str(v.value) for k,v in self.entries.items()},indent=2, sort_keys=True)

    def debug_repr(self) -> str:
        """
        repr() intended for debug purposes, since it runs repr() on each entry.value,
        it is more accurate than str(entry.value)
        """
        repr_string = "{}(Confi):\n".format(self.__class__.__name__)
        items = list(self.entries.items())
        items.sort(key = lambda item: item[0])
        indent = ' ' * 4
        for key, entry in items:
            repr_string += f"{indent}{key}: {repr(entry.value)}\n"
        return repr_string

    def get_cli_object(self, config_objects:List["Confi"]=None, typer_app:Typer=None,
                       help:str=None, on_start:Callable=None):

        if config_objects is None:
            config_objects = []
        config_objects.append(self)
        return get_cli_object_for_config_objects(config_objects, typer_app=typer_app, help=help, on_start=on_start)

    def cli(self,config_objects:List["Confi"]=None, typer_app:Typer=None, help:str=None, on_start:Callable=None):
        """
        Run a command-line-interface based on this configuration set, other config sets, and s typer cli app

        Args:
            config_objects (List[Confi, optional): additional config objects to share the CLI with this one. Defaults to None.
            typer_app (Typer, optional): A typer cli app with commands to expose to the CLI. Defaults to None.
        """
        self.get_cli_object(config_objects, typer_app=typer_app, help=help, on_start=on_start)()


    def on_load(self):
        """
        Callback called upon configuration load
        Add dynamic values you want set here (i.e. values which are based on other values)
        """
        pass

    def __setattr__(self, name: str, value: Any) -> None:
        """
            Make sure value updates are saved in internal entries as well
        """
        super().__setattr__(name, value)
        # update entry as well (to sync with CLI, etc. )
        if not name.startswith("_") and name in self._entries:
            self._entries[name].value = value

    def delay(self, value):
        delayed_entry = ConfiDelay(value, index=self._counter)
        self._counter += 1
        return delayed_entry

    # -- parser setters --

    def str(self, key, default=undefined, description=None, **kwargs) -> str:
        return self._process(key, description=description, default=default, type=str, **kwargs)

    def int(self, key, default=undefined, description=None, **kwargs) -> int:
        return self._process(key, description=description, default=default, cast=int, type=int, **kwargs)

    def bool(self, key, default=undefined, description=None, **kwargs) -> bool:
        return self._process(key, description=description, default=default, cast=cast_boolean, type=bool, **kwargs)

    def float(self, key, default=undefined, description=None, **kwargs) -> float:
        return self._process(key, description=description, default=default, cast=float, type=float, **kwargs)

    def list(self, key, default=undefined, sub_cast=text_type, delimiter=",", strip=string.whitespace, description=None, **kwargs) -> list:
        return self._process(key, default=default, description=description, cast=Csv(cast=sub_cast, delimiter=delimiter, strip=strip), type=list, **kwargs)

    def model(self, key, model_type:T, default=undefined, description=None, **kwargs) -> T:
        """
        Parse a config using a Pydantic model
        """
        return self._process(key, description=description, default=default, cast=cast_pydantic(model_type),type=model_type, **kwargs)

    def enum(self, key, enum_type: EnumT, default=undefined, description=None, **kwargs) -> EnumT:
        return self._process(key, description=description, default=default, cast=enum_type, type=enum_type, **kwargs)

    def private_key(
        self,
        key: str,
        default: Any = undefined,
        description: str = None,
        key_format: Optional[EncryptionKeyFormat] = None,
        passphrase: Optional[str] = None,
        **kwargs
    ) -> Optional[PrivateKey]:
        """
        parse a cryptographic private key from env vars
        """
        cast_key = partial(cast_private_key, key_format=key_format, passphrase=passphrase)
        return self._process(key, description=description, default=default, cast=cast_key,type=PrivateKey, **kwargs)

    def public_key(
        self,
        key: str,
        default: Any = undefined,
        description: str = None,
        key_format: Optional[EncryptionKeyFormat] = None,
        **kwargs
    ) -> Optional[PublicKey]:
        """
        parse a cryptographic public key from env vars
        """
        cast_key = partial(cast_public_key, key_format=key_format)
        return self._process(key, description=description, default=default, cast=cast_key, type=PublicKey, **kwargs)


# default parser
confi = Confi()





import pytest
import os
import sys

# Add root opal dir to use local src as package for tests (i.e, no need for python -m pytest)
root_dir = os.path.abspath(
    os.path.join(
        os.path.dirname(__file__),
        os.path.pardir,
        os.path.pardir,
        os.path.pardir,
    )
)
sys.path.append(root_dir)

# -----------------------------------------------------------------------------
import asyncio
from typing import List, Optional
from pathlib import Path
from uuid import uuid4
from datetime import timedelta

from opal_common.authentication.types import EncryptionKeyFormat, JWTAlgorithm, JWTClaims, PrivateKey, PublicKey
from opal_common.authentication.casting import cast_private_key, cast_public_key
from opal_common.authentication.signer import JWTSigner
from opal_common.authentication.verifier import JWTVerifier

KEY_FILENAME = "opal_test_crypto_key"
PASSPHRASE = "whiterabbit"

AUTH_JWT_AUDIENCE = "https://api.authorizon.com/v1/"
AUTH_JWT_ISSUER = "https://authorizon.com/"
CUSTOM_CLAIMS = {"color": "red"}


async def run_subprocess(command: str):
    process = await asyncio.create_subprocess_shell(
        command,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        start_new_session=True,
    )
    return_code = await process.wait()
    assert return_code == 0

async def run_commands(commands: List[str]):
    # the await inside the for-loop is intentional, these commands should not run in parallel
    for command in commands:
        await run_subprocess(command)

async def verify_crypto_keys(
    private_key_filename: Path,
    public_key_filename: Path,
    private_key_format: EncryptionKeyFormat,
    public_key_format: EncryptionKeyFormat,
    algorithm: JWTAlgorithm,
    passphrase: Optional[str] = None
    ):
    # assert keys created
    assert private_key_filename.exists()
    assert public_key_filename.exists()

    private_key: Optional[PrivateKey] = cast_private_key(private_key_filename, private_key_format, passphrase)
    assert private_key is not None

    public_key: Optional[PublicKey] = cast_public_key(public_key_filename, public_key_format)
    assert public_key is not None

    verifier = JWTVerifier(public_key, algorithm, AUTH_JWT_AUDIENCE, AUTH_JWT_ISSUER)
    assert verifier.enabled

    signer = JWTSigner(private_key, public_key, algorithm, AUTH_JWT_AUDIENCE, AUTH_JWT_ISSUER)
    assert signer.enabled

    token: str = signer.sign(uuid4(), timedelta(days=1), CUSTOM_CLAIMS)
    claims: JWTClaims = verifier.verify(token)

    for k in CUSTOM_CLAIMS:
        assert k in claims.keys()
        assert CUSTOM_CLAIMS[k] == claims[k]


@pytest.mark.asyncio
async def test_encryption_keys_RFC_4253_ssh_format_with_passphrase(tmp_path):
    """
    Test key encryption format: RFC_4253

    such keys can be generate by this command:
    ```
    ssh-keygen -t rsa -b 4096 -m pem
    ```
    the private key is in PEM format
    the public key in in ssh format
    """
    # creates the keys under temp paths that are auto-deleted after the test
    private_key_filename = Path(os.path.join(tmp_path, KEY_FILENAME))
    public_key_filename = Path(f"{private_key_filename}.pub")

    # commands to generate crypto keys
    await run_commands([
        f"ssh-keygen -t rsa -b 4096 -m pem -f {private_key_filename} -N {PASSPHRASE}",
    ])

    await verify_crypto_keys(
        private_key_filename = private_key_filename,
        public_key_filename = public_key_filename,
        private_key_format = EncryptionKeyFormat.pem,
        public_key_format = EncryptionKeyFormat.ssh,
        algorithm = getattr(JWTAlgorithm, "RS256"),
        passphrase= PASSPHRASE
    )

@pytest.mark.asyncio
async def test_encryption_keys_RFC_4253_ssh_format_no_passphrase(tmp_path):
    # creates the keys under temp paths that are auto-deleted after the test
    private_key_filename = Path(os.path.join(tmp_path, KEY_FILENAME))
    public_key_filename = Path(f"{private_key_filename}.pub")

    # commands to generate crypto keys
    await run_commands([
        f"ssh-keygen -t rsa -b 4096 -m pem -f {private_key_filename}",
    ])

    await verify_crypto_keys(
        private_key_filename = private_key_filename,
        public_key_filename = public_key_filename,
        private_key_format = EncryptionKeyFormat.pem,
        public_key_format = EncryptionKeyFormat.ssh,
        algorithm = getattr(JWTAlgorithm, "RS256"),
        passphrase = None
    )

@pytest.mark.asyncio
async def test_encryption_keys_PKCS1_format_with_passphrase(tmp_path):
    # creates the keys under temp paths that are auto-deleted after the test
    private_key_filename = Path(os.path.join(tmp_path, KEY_FILENAME))
    public_key_filename = Path(f"{private_key_filename}.pub")

    # commands to generate crypto keys
    await run_commands([
        f"ssh-keygen -t rsa -b 4096 -m pem -f {private_key_filename} -N {PASSPHRASE}",
        f"ssh-keygen -e -m pem -f {private_key_filename} -P {PASSPHRASE} > {public_key_filename}"
    ])

    await verify_crypto_keys(
        private_key_filename = private_key_filename,
        public_key_filename = public_key_filename,
        private_key_format = EncryptionKeyFormat.pem,
        public_key_format = EncryptionKeyFormat.pem,
        algorithm = getattr(JWTAlgorithm, "RS256"),
        passphrase = PASSPHRASE
    )

@pytest.mark.asyncio
async def test_encryption_keys_X509_SPKI_format_with_passphrase(tmp_path):
    # creates the keys under temp paths that are auto-deleted after the test
    private_key_filename = Path(os.path.join(tmp_path, KEY_FILENAME))
    public_key_filename = Path(f"{private_key_filename}.pub")

    # commands to generate crypto keys
    await run_commands([
        f"ssh-keygen -t rsa -b 4096 -m pem -f {private_key_filename} -N {PASSPHRASE}",
        f"ssh-keygen -e -m pkcs8 -f {private_key_filename} -P {PASSPHRASE} > {public_key_filename}"
    ])

    await verify_crypto_keys(
        private_key_filename = private_key_filename,
        public_key_filename = public_key_filename,
        private_key_format = EncryptionKeyFormat.pem,
        public_key_format = EncryptionKeyFormat.pem,
        algorithm = getattr(JWTAlgorithm, "RS256"),
        passphrase = PASSPHRASE
    )

@pytest.mark.asyncio
async def test_encryption_keys_PKCS1_format_with_passphrase_hardcoded_keys(tmp_path):
    """
    these hardcoded keys are for test purposes only - NEVER use them!!!!
    """
    # creates the keys under temp paths that are auto-deleted after the test
    private_key_filename = Path(os.path.join(tmp_path, KEY_FILENAME))
    public_key_filename = Path(f"{private_key_filename}.pub")

    open(private_key_filename, 'w').write("""
-----BEGIN RSA PRIVATE KEY-----
Proc-Type: 4,ENCRYPTED
DEK-Info: AES-128-CBC,76065A0C88ABA1D6B5B812ED6BE346B5
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-----END RSA PRIVATE KEY-----"""
    )

    open(public_key_filename, 'w').write("""
-----BEGIN RSA PUBLIC KEY-----
MIICCgKCAgEAoTtleKwCPQ4kUGAAk4uDnjZStr/LbzFaYTPrfFWbRmEFzVoWREST
STesIuoTgdLLiDY0wksC6IZ/wlvVvxbuRG4PljUPiOvfHTDI2giC7JoH40b0HNI+
1YlDWU+pnvS1SeoPzzVwC2uVP97g0Z0d7BlXZY8ufsvQPfHQP4sTuvFXiMNC18dQ
OVLZS8EpeNhORf+iTe43OgNNnD4KCl9w1VScUlR58K+N+glQ6qONZjayII9Qn0vP
IZfw/RseR5qprJuA30s+hPMtOhqwfoWwIP8dGRL/rbPoVb0tvLTWnmbDySQGDyof
7aWlW0LN+7/5wPlxhq8coaRcTxlCU0Iro1xL/TpOvOQb4Q+++qUphIc+t1a8j2Ge
X4/+N5kSSNU83NJpEGLLknaoiR9cYvNBJksJhAGHvu0kLmb5/91nK/CHJMsma82W
GR9TQAtoSOlOh8P3cpeLCEHdMd3iXHdMUQ7rzIRGjZyPPMv+JxumQ1MJ76vNHB4j
JgN5LRWeo+H17t7bnO/Ot6qmmp7ZN3dc3QBlsy09cCdQ4l4YWEa3VO6dXnIdoe/c
THdYobue6ft5E7d7Eez4is6++d8SuboxJMmzbDK9U++GoJVARk2FqUpXTqHSIqKt
y8eFhYJfV0a59E5TasHlIT/HLcdlvISQ0/lBoPOwtKDbFuZvqQwWKEsCAwEAAQ==
-----END RSA PUBLIC KEY-----
    """)

    await verify_crypto_keys(
        private_key_filename = private_key_filename,
        public_key_filename = public_key_filename,
        private_key_format = EncryptionKeyFormat.pem,
        public_key_format = EncryptionKeyFormat.pem,
        algorithm = getattr(JWTAlgorithm, "RS256"),
        passphrase = PASSPHRASE
    )

@pytest.mark.asyncio
async def test_encryption_keys_X509_SPKI_format_with_passphrase_hardcoded_keys(tmp_path):
    """
    these hardcoded keys are for test purposes only - NEVER use them!!!!
    """
    # creates the keys under temp paths that are auto-deleted after the test
    private_key_filename = Path(os.path.join(tmp_path, KEY_FILENAME))
    public_key_filename = Path(f"{private_key_filename}.pub")

    open(private_key_filename, 'w').write("""
-----BEGIN RSA PRIVATE KEY-----
Proc-Type: 4,ENCRYPTED
DEK-Info: AES-128-CBC,B0E9701EF36B6D18081BE93E446F7DAC
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-----END RSA PRIVATE KEY-----"""
    )

    open(public_key_filename, 'w').write("""
-----BEGIN PUBLIC KEY-----
MIICIjANBgkqhkiG9w0BAQEFAAOCAg8AMIICCgKCAgEA3dPvy1zGdxMRf06+KYbk
/OW73J8t1eX4Nb3E5w5GV6gdt1KKFMD7WOy9wPFNUS1k9o7WhW1giQGSIE1NuShH
Y8x3yvYnPHVptrWFmKZMTVb8jytmEtLT0tHfJjCJKjPUqUT2Xli/Lyg/MHz0zqPb
b+vlXFBujI+3u1em6DUMP5Qao3rq0qzdD2fmBVsXrcOhAL9GwgyHTFVON/PvBKXU
MvseZ+FAqydBdG95b+DKG4eEzjz0CV3SXhMx4orPUsa9VupAbve3MMGYyXB51yYh
lDyTnFiAChhIzU99pgg5Uhx8YX/hFEAoATwbv6pmliFmUzuTs5sLP9DP+56FGye8
s1wxIMBMAMzBTMKnsJlpIOyEBq8tI0UiSM76jP6cAdK/JgZ+tHi2HJXbtm+Gc02F
2syHJI7lryQNfbPqgn8QXJNVMiCSnmqmXfEoz2cWkqTn9siIovnrOsQDN4nfiLGs
tjxiX307tP6CjNFl24/olZdan6D4jxskI1cgXxL53tjdXASbJ5Z3M/xJmm5P/eGv
GKe4LpzupEUoDz4UnfgPUsOORe4p49NHm07S3KCouZOAslMAJDqe2qyyPhDnUaTG
0OAmlqa412uGUhmxYbKgyPNALED/9WksFEWDDqt+bq6zwYActCW2203gCRsTbSX9
wo0Src+YUGAdjomgzrt/6CECAwEAAQ==
-----END PUBLIC KEY-----
    """)

    await verify_crypto_keys(
        private_key_filename = private_key_filename,
        public_key_filename = public_key_filename,
        private_key_format = EncryptionKeyFormat.pem,
        public_key_format = EncryptionKeyFormat.pem,
        algorithm = getattr(JWTAlgorithm, "RS256"),
        passphrase = PASSPHRASE
    )


"""
Utilities to run UVICORN / GUNICORN
"""
import multiprocessing
from typing import Dict
import gunicorn.app.base

def calc_default_number_of_workers():
    return (multiprocessing.cpu_count() * 2) + 1

class GunicornApp(gunicorn.app.base.BaseApplication):

    def __init__(self, app, options:Dict[str,str]=None):
        self.options = options or {}
        self.application = app
        super().__init__()

    def load_config(self):
        config = {key: value for key, value in self.options.items()
                  if key in self.cfg.settings and value is not None}
        for key, value in config.items():
            self.cfg.set(key.lower(), value)

    def load(self):
        return self.application


def run_gunicorn(app, number_of_workers=None,host=None, port=None, **kwargs):
    options = {
        'bind': '%s:%s' %  (host or '127.0.0.1', port or '8080'),
        'workers': number_of_workers or calc_default_number_of_workers(),
        'worker_class': 'uvicorn.workers.UvicornWorker'
    }
    options.update(kwargs)
    GunicornApp(app, options).run()


def run_uvicorn(app_path, number_of_workers=None,host=None, port=None, reload=False, **kwargs):
    options = {
        'host': host or '127.0.0.1',
        'port': port or '8080',
        'reload': reload,
        'workers': number_of_workers or calc_default_number_of_workers(),
    }
    options.update(kwargs)
    import uvicorn
    uvicorn.run(app_path, **options)
    



from functools import partial
from pathlib import Path
from typing import List, Optional, Set

from git import Repo
from git.objects import Commit

from opal_common.paths import PathUtils
from opal_common.git.commit_viewer import CommitViewer, VersionedFile, has_extension, is_under_directories
from opal_common.git.diff_viewer import DiffViewer, diffed_file_has_extension, diffed_file_is_under_directories
from opal_common.opa import get_rego_package, is_data_module, is_rego_module
from opal_common.schemas.policy import DataModule, PolicyBundle, RegoModule, DeletedFiles


class BundleMaker:
    """
    creates a policy bundle based on:
    - the current state of the policy git repo
    - filtering criteria on the policy git repo (specific directories, specific file types, etc)

    there are two types of bundles:
    - a full/complete bundle, representing the state of the repo at one commit
    - a diff bundle, representing only the *changes* made to the policy between two commits (the diff).
    """
    def __init__(self, repo: Repo, in_directories: Set[Path], extensions: Optional[List[str]] = None, manifest_filename: str = ".manifest"):
        """[summary]

        Args:
            repo (Repo): the policy repo
            in_directories (Set[Path]): the directories in the repo that we want to filter on.
                if the entire repo is relevant, pass Path(".") as the directory
                (all paths are relative to the repo root).
            extensions (Optional[List[str]]): optional filtering on file extensions.
        """
        self._repo = repo
        self._directories = in_directories
        self._has_extension = partial(has_extension, extensions=extensions)
        self._is_under_directories = partial(is_under_directories, directories=in_directories)
        self._diffed_file_has_extension = partial(diffed_file_has_extension, extensions=extensions)
        self._diffed_file_is_under_directories = partial(diffed_file_is_under_directories, directories=in_directories)
        self._manifest_filename = manifest_filename

    def _get_explicit_manifest(self, viewer: CommitViewer) -> Optional[List[str]]:
        """
        Rego policies often have dependencies (import statements) between policies.
        Since the OPAL client is limited by the OPA REST api and this api currently does not allow
        to load bundles (multiple policies together), OPAL client can only load one policy at a time.

        If policies with dependencies between them are loaded out-of-order, OPA will throw an exception.

        To mitigate this, we allow the developer to put a manifest file in the repository (default path: .manifest).
        This file, if exists, should contain the list of policy files (.rego) in the correct order they should be
        loaded into OPA.

        This method searches for an explicit manifest file, reads it and returns the list of paths, or None if not found.
        """
        def is_manifest_file(f: VersionedFile) -> bool:
            return f.path == Path(self._manifest_filename)

        manifest_files = list(viewer.files(is_manifest_file))
        if not manifest_files:
            return None

        try:
            manifest_paths = [Path(path) for path in manifest_files[0].read().splitlines()]
            return [str(path) for path in manifest_paths if viewer.exists(path)]
        except:
            return None

    def _sort_manifest(self, unsorted_manifest: List[str], explicit_sorting: Optional[List[str]]) -> List[str]:
        """
        the way this sorting works, is assuming that explicit_sorting does NOT necessarily contains all the policies
        found in the manifest, or that even "policies" mentioned in it actually exists in the actual generated manifest.
        We must ensure that all items in unsorted_manifest must also exist in the output list.
        """
        if not explicit_sorting:
            return unsorted_manifest

        # casting to Path
        unsorted_paths = [Path(path) for path in unsorted_manifest]
        sorting = [Path(path) for path in explicit_sorting]

        # sorting the list
        sorted_paths = PathUtils.sort_paths_according_to_explicit_sorting(unsorted_paths, sorting)

        # cast back to string paths
        return [str(path) for path in sorted_paths]

    def make_bundle(self, commit: Commit) -> PolicyBundle:
        """
        creates a *complete* bundle of all the policy and data modules found
        in the policy repo, when the repo HEAD is at the given `commit`.

        Args:
            commit (Commit): the commit the repo should be checked out on to search for policy files.

        Returns:
            bundle (PolicyBundle): the bundle of policy modules found in the repo (checked out on `commit`)
        """
        data_modules = []
        policy_modules = []
        manifest = []

        with CommitViewer(commit) as viewer:
            filter = lambda f: self._has_extension(f) and self._is_under_directories(f)
            explicit_manifest = self._get_explicit_manifest(viewer)
            for source_file in viewer.files(filter):
                contents = source_file.read()
                path = source_file.path

                if is_data_module(path):
                    data_modules.append(
                        DataModule(path=str(path.parent), data=contents)
                    )
                elif is_rego_module(path):
                    policy_modules.append(
                        RegoModule(
                            path=str(path),
                            package_name=get_rego_package(contents) or "",
                            rego=contents,
                        )
                    )
                else:
                    continue

                manifest.append(str(path)) # only returns the relative path

            return PolicyBundle(
                manifest=self._sort_manifest(manifest, explicit_manifest),
                hash=commit.hexsha,
                data_modules=data_modules,
                policy_modules=policy_modules,
            )

    def make_diff_bundle(self, old_commit: Commit, new_commit: Commit) -> PolicyBundle:
        """
        creates a *diff* bundle of all the policy and data modules that were changed
        (either added, renamed, modified or deleted) between the `old_commit` and `new_commit`.
        essentially all the relevant files when running `git diff old_commit..new_commit`.

        Note that we still filter only directories and file types given in the constructor.

        Args:
            old_commit (Commit): represents the previous known state of the repo.
                The opal client subscribes to the policy state and gets updates from
                the server via a pubsub channel. When it receives an update that
                new state is available in the policy repo, the client requests a
                *diff bundle* from the /policy api route. The client will report
                its last known commit as the `old_commit`, and only new state
                (the diff from the client known commit to the server newest commit)
                will be returned back.
            commit (Commit): represents the newest known commit in the server (the new state).

        Returns:
            bundle (PolicyBundle): a diff bundle containing only the policy modules changed
                between `old_commit` and `new_commit`.
        """
        data_modules = []
        policy_modules = []
        deleted_data_modules = []
        deleted_policy_modules = []
        manifest = []
        explicit_manifest = self._get_explicit_manifest(CommitViewer(new_commit))

        with DiffViewer(old_commit, new_commit) as viewer:
            filter = lambda diff: (
                self._diffed_file_has_extension(diff) and \
                    self._diffed_file_is_under_directories(diff)
                )
            for source_file in viewer.added_or_modified_files(filter):
                contents = source_file.read()
                path = source_file.path

                if is_data_module(path):
                    data_modules.append(
                        # in OPA, the data module path is the containing directory
                        # i.e: /path/to/data.json will put the json contents under "/path/to" in the opa tree
                        DataModule(path=str(path.parent), data=contents)
                    )
                elif is_rego_module(path):
                    policy_modules.append(
                        RegoModule(
                            path=str(path),
                            package_name=get_rego_package(contents) or "",
                            rego=contents,
                        )
                    )
                else:
                    continue

                manifest.append(str(path)) # only returns the relative path

            for source_file in viewer.deleted_files(filter):
                path = source_file.path

                if is_data_module(path):
                    # in OPA, the data module path is the containing directory (see above)
                    deleted_data_modules.append(str(path.parent))
                elif is_rego_module(path):
                    deleted_policy_modules.append(path)

            if deleted_data_modules or deleted_policy_modules:
                deleted_policies = self._sort_manifest(deleted_policy_modules, explicit_manifest)
                deleted_policies.reverse() # when removed, dependent policies should be removed first

                deleted_files = DeletedFiles(
                    data_modules=deleted_data_modules,
                    policy_modules=deleted_policies,
                )
            else:
                deleted_files = None

            return PolicyBundle(
                manifest=self._sort_manifest(manifest, explicit_manifest),
                hash=new_commit.hexsha,
                old_hash=old_commit.hexsha,
                data_modules=data_modules,
                policy_modules=policy_modules,
                deleted_files=deleted_files
            )



from typing import List, Dict, Any
from pydantic import BaseModel, Field
from datetime import datetime
from enum import Enum


class TransactionType(str, Enum):
    policy = 'policy'
    data = 'data'


class RemoteStatus(BaseModel):
    remote_url: str = Field(None, description="Url of remote data/policy source")
    succeed: bool = Field(True, description="Is request succeed")
    error: str = Field(None, description="If failed contains the type of exception")


class StoreTransaction(BaseModel):
    """
    represents a transaction of policy or data to OPA
    """
    id: str = Field(..., description="The id of the transaction")
    actions: List[str] = Field(..., description="The write actions performed as part of the transaction")
    transaction_type: TransactionType = Field(None, description="Type of transaction,is it data/policy transaction")
    success: bool = Field(False, description="Whether or not the transaction was successful")
    error: str = Field("", description="Error message in case of failure, defaults to empty string")
    creation_time: str = Field(None, description="Creation time for this store transaction")
    end_time: str = Field(None, description="Finish time for this store transaction")
    remotes_status: List[RemoteStatus] = Field(None, description="List of the remote sources for this transaction and their status")


class JSONPatchAction(BaseModel):
    """
    Abstract base class for JSON patch actions (RFC 6902)
    """
    op: str = Field(..., description="patch action to perform")
    path: str = Field(..., description="target location in modified json")
    value: Dict[str, Any] = Field(..., description="json document, the operand of the action")

class ArrayAppendAction(JSONPatchAction):
    op: str = Field("add", description="add action -> adds to the array")
    path: str = Field("-", description="dash marks the last index of an array")


from pathlib import Path
from fastapi.exceptions import HTTPException
from tenacity import AsyncRetrying
from typing import Optional, Tuple
import aiohttp
from datetime import datetime
from fastapi import status
from tenacity.wait import wait_fixed
from opal_client.policy.fetcher import throw_if_bad_status_code
from opal_client.utils import tuple_to_dict
from opal_common.git.tar_file_to_local_git_extractor import TarFileToLocalGitExtractor

from opal_common.logger import logger
from opal_common.sources.base_policy_source import BasePolicySource
from opal_common.utils import get_authorization_header, hash_file
from opal_server.config import opal_server_config

BundleHash = str


class ApiPolicySource(BasePolicySource):
    """
    Watches an OPA-like bundle server for changes and can trigger callbacks
    when detecting a new bundle.

    Checking for changes is done by sending an HTTP GET request to the remote bundle server.
    OPAL will check for changes either when triggered a webhook or periodically if configured
    to run a polling task.

    You can read more on OPA bundles here:
    https://www.openpolicyagent.org/docs/latest/management-bundles/

    Args:
        remote_source_url(str): the base address to request the policy from
        local_clone_path(str):  path for the local git to manage policies
        polling_interval(int):  how many seconds need to wait between polling
        token (str, optional):  auth token to include in connections to OPAL server. Defaults to POLICY_BUNDLE_SERVER_TOKEN.
    """

    def __init__(
        self,
        remote_source_url: str,
        local_clone_path: str,
        polling_interval: int = 0,
        token: Optional[str] = None,
    ):
        super().__init__(remote_source_url=remote_source_url, local_clone_path=local_clone_path,
                         polling_interval=polling_interval)
        self.token = token
        self.bundle_hash = None
        self.etag = None
        self.tmp_bundle_path = Path(opal_server_config.POLICY_BUNDLE_TMP_PATH)
        self.tar_to_git = TarFileToLocalGitExtractor(self.local_clone_path, self.tmp_bundle_path)

    async def get_initial_policy_state_from_remote(self):
        """
        init remote data to local repo
        """
        async for attempt in AsyncRetrying(wait=wait_fixed(5)):
            with attempt:
                await self.fetch_policy_bundle_from_api_source(self.remote_source_url, self.token)
                self.local_git = self.tar_to_git.create_local_git()

    async def api_update_policy(self) -> Tuple[bool, str, str]:
        async for attempt in AsyncRetrying(wait=wait_fixed(5)):
            with attempt:
                tmp_bundle_path, prev_version, current_hash = await self.fetch_policy_bundle_from_api_source(self.remote_source_url, self.token)
                if tmp_bundle_path and prev_version and current_hash:
                    commit_msg = f"new version {current_hash}"
                    self.local_git, prev_commit, new_commit = self.tar_to_git.extract_bundle_to_local_git(commit_msg=commit_msg)
                    return True, prev_version, current_hash, prev_commit, new_commit
                else:
                    return False, None, current_hash, None, None

    async def fetch_policy_bundle_from_api_source(
        self,
        url: str,
        token: Optional[str]
    ) -> Tuple[Path, BundleHash, BundleHash]:
        """
        Fetches the bundle. May throw, in which case we retry again.
        Checks that the bundle file isn't the same with Etag, if server
        doesn't have Etag it checks it with hash on the bundle file

        Read more on Etag here:
        https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/ETag

        Args:
            url(str): the base address to request the bundle.tar.gz file from
            token (str, optional): Auth token to include in connections to OPAL server. Defaults to POLICY_BUNDLE_SERVER_TOKEN.
        Returns:
            Path: path to the bundle file that we just downloaded from the remote API source
            BundleHash: previous bundle hash on None if this is the initial bundle file
            BundleHash: current bundle hash

        """
        auth_headers = tuple_to_dict(
            get_authorization_header(token)) if token else {}
        etag_headers = {'ETag': self.etag,
                        "If-None-Match": self.etag} if self.etag else {}
        full_url = f"{url}/bundle.tar.gz"

        async with aiohttp.ClientSession() as session:
            try:
                async with session.get(
                    f"{full_url}",
                    headers={'content-type': 'application/gzip',
                             **auth_headers, **etag_headers},
                ) as response:
                    if response.status == status.HTTP_404_NOT_FOUND:
                        logger.warning(
                            "requested url not found: {full_url}", full_url=full_url)
                        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=f"requested url not found: {full_url}")
                    if response.status == status.HTTP_304_NOT_MODIFIED:
                        logger.info(
                            "Not modified at: {now}", now=datetime.now().strftime('%Y-%m-%d %H:%M:%S'))
                        return False, None, self.etag

                    # may throw ValueError
                    await throw_if_bad_status_code(response, expected=[status.HTTP_200_OK])
                    current_etag = response.headers.get("ETag", None)
                    response_bytes = await response.read()
                    tmp_file_path = Path(opal_server_config.POLICY_BUNDLE_TMP_PATH)
                    with open(tmp_file_path, "wb") as file:
                        file.write(response_bytes)

                    if not current_etag:
                        logger.info('Etag is turnned off, you may want to turn it on at your bundle server')
                        current_bundle_hash = hash_file(tmp_file_path)
                        logger.info('Bundle hash is {hash}', hash=current_bundle_hash)
                        if self.bundle_hash == current_bundle_hash:
                            logger.info("No new bundle, hash is: {hash}", hash=current_bundle_hash)
                            return False, None, current_bundle_hash
                        else:
                            logger.info("New bundle found, hash is: {hash}", hash=current_bundle_hash)
                            prev_bundle_hash = self.bundle_hash
                            self.bundle_hash = current_bundle_hash
                            return tmp_file_path, prev_bundle_hash, current_bundle_hash
                    else:
                        if self.etag == current_etag: # validate against bad etag implementation
                            logger.info("No new bundle, hash is: {hash}", hash=current_etag)
                            return False, None, current_etag
                        prev_etag = self.etag
                        self.etag = current_etag
                        return tmp_file_path, prev_etag, current_etag

            except (aiohttp.ClientError, HTTPException) as e:
                logger.warning("server connection error: {err}", err=repr(e))
                raise
            except Exception as e:
                logger.error("unexpected server connection error: {err}", err=repr(e))
                raise

    async def check_for_changes(self):
        """
        Calling this method will trigger an api check to the remote.
        If after the request the watcher detects new bundle, it will call the
        callbacks registered with _on_new_policy().
        """
        logger.info(
            "Fetching changes from remote: '{remote}'", remote=self.remote_source_url)
        has_changes, prev, latest, prev_commit, new_commit = await self.api_update_policy()
        if not has_changes:
            logger.info(
                "No new version: current hash is: {head}", head=latest)
        else:
            logger.info(
                "Found new version: old version hash was '{prev_head}', new version hash is '{new_head}'",
                prev_head=prev, new_head=latest)
            await self._on_new_policy(old=prev_commit, new=new_commit)



import logging
import os
import sys

# Add parent path to use local src as package for tests
root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir, os.path.pardir, os.path.pardir ))
sys.path.append(root_dir)

import asyncio
from multiprocessing import Process

import pytest
import uvicorn
from fastapi import FastAPI
from fastapi_websocket_rpc import WebsocketRPCEndpoint, RpcMethodsBase


from opal_common.fetcher import FetchingEngine
from opal_common.fetcher.providers.fastapi_rpc_fetch_provider import FastApiRpcFetchProvider, FastApiRpcFetchEvent, FastApiRpcFetchConfig


# Configurable
PORT = int(os.environ.get("PORT") or "9110")
uri = f"ws://localhost:{PORT}/rpc"
DATA_PREFIX = "I AM DATA - HEAR ME ROAR"
SUFFIX = " - Magic!"


class RpcData(RpcMethodsBase):

    async def get_data(self, suffix: str) -> str:
        return DATA_PREFIX + suffix


def setup_server():
    app =  FastAPI()
    endpoint = WebsocketRPCEndpoint(RpcData())
    endpoint.register_route(app,"/rpc")

    uvicorn.run(app, port=PORT )

@pytest.fixture(scope="module")
def server():
    # Run the server as a separate process
    proc = Process(target=setup_server, args=(), daemon=True)
    proc.start()
    yield proc
    proc.kill() # Cleanup after test

@pytest.mark.asyncio
async def test_simple_rpc_fetch(server):
    """
    """
    got_data_event = asyncio.Event()
    async with FetchingEngine() as engine:
        engine.register.register_fetcher(FastApiRpcFetchProvider.__name__, FastApiRpcFetchProvider)
        # Event for RPC fetch
        fetch_event = FastApiRpcFetchEvent(url=uri, config=FastApiRpcFetchConfig(rpc_method_name="get_data", rpc_arguments={"suffix": SUFFIX}))
        # Callback for event
        async def callback(result):
            data = result.result
            assert data == DATA_PREFIX + SUFFIX
            got_data_event.set()
        await engine.queue_fetch_event(fetch_event, callback)
        await asyncio.wait_for(got_data_event.wait(), 5)
        assert got_data_event.is_set()




import os
from pathlib import Path
import shutil
from typing import List, Optional
import git
from pydantic.error_wrappers import ValidationError

from opal_common.security.tarsafe import TarSafe
from opal_server.config import opal_server_config


class TarFileToLocalGitExtractor:
    """
    This class takes tar file from remote api source and extract it to local
    git, so we could manage update to opal clients

    Args:
        local_clone_path(str):  path for the local git to manage policies
        tmp_bundle_path(Path):  path to download bundle from api source
    """
    def __init__(self, local_clone_path: str, tmp_bundle_path: Path):
        self.local_clone_path = local_clone_path
        self.tmp_bundle_path = tmp_bundle_path

    def commit_local_git(self, init_commit_msg: str = "Init", should_init: bool = False):
        """
        Commit first version of bundle or the updates that come after
        Args:
            init_commit_msg(str):  text of the commit msg
            should_init(Path):  should it init the repo or it is existing repo
        """
        if should_init:
            local_git = git.Repo.init(self.local_clone_path)
        else:
            local_git = git.Repo(self.local_clone_path)
        prev_commit = None
        if len(local_git.index.repo.heads):
            prev_commit = local_git.index.repo.head.commit
        local_git.index.add(opal_server_config.POLICY_BUNDLE_GIT_ADD_PATTERN)
        new_commit = local_git.index.commit(init_commit_msg)
        return local_git, prev_commit, new_commit

    def create_local_git(self):
        """
        Extract bundle create local git and commit this initial state
        """

        self.extract_bundle_tar()
        local_git = TarFileToLocalGitExtractor.is_git_repo(self.local_clone_path)
        if not local_git or len(local_git.heads) == 0:
            local_git = self.commit_local_git(should_init=True)
        return local_git

    def extract_bundle_to_local_git(self, commit_msg: str):
        """
        Update local git with new bundle
        Args:
            commit_msg(str):  text of the commit msg
        """
        tmp_path = f"{self.local_clone_path}.bak"
        os.rename(self.local_clone_path, tmp_path)
        try:
            self.extract_bundle_tar()
            shutil.move(os.path.join(tmp_path, ".git"), os.path.join(self.local_clone_path, ".git"))
        finally:
            shutil.rmtree(tmp_path)
        local_git, prev_commit, new_commit = self.commit_local_git(commit_msg)
        return local_git, prev_commit, new_commit

    def extract_bundle_tar(self, mode: str = "r:gz") -> bool:
        """
        Extract bundle tar, tar path is at self.tmp_bundle_path
        Uses TarSafe that checks that our bundle file don't have vulnerabilities like path traversal
        Args:
            mode(str): mode for TarSafe default to r:gz that can open tar.gz files
        """
        with TarSafe.open(self.tmp_bundle_path, mode=mode) as tar_file:
            tar_file_names = tar_file.getnames()
            TarFileToLocalGitExtractor.validate_tar_or_throw(tar_file_names)
            tar_file.extractall(path=self.local_clone_path)

    @staticmethod
    def is_git_repo(path) -> Optional[git.Repo]:
        """
        Checks is this path is a git repo if it is return Repo obj
        Return:
            Repo obj if it is a git repo if not returns None
        """
        local_git = False
        try:
            local_git = git.Repo(path)
            _ = local_git.git_dir
            return local_git
        except Exception:
            return None

    @staticmethod
    def validate_tar_or_throw(tar_file_names: List[str], forbidden_filename: str = '.git'):
        if len(tar_file_names) == 0:
            raise ValidationError("No files in bundle")
        if forbidden_filename and forbidden_filename in tar_file_names:
            raise ValidationError("No {forbidden_filename} files are allowed in OPAL api bundle".format(
                forbidden_filename=forbidden_filename))



class GitFailed(Exception):
    """
    an exception we throw on git failures that are caused by wrong assumptions.
    i.e: we want to track a non-existing branch, or git url is not valid.
    """
    def __init__(self, exc: Exception):
        self._original_exc = exc
        super().__init__()


from ..events import FetchEvent

# Callback signatures 
async def OnFetchFailureCallback(exception:Exception, event: FetchEvent):
    """
    Args:
        exception (Exception): The exception thrown causing the failure 
        event (FetchEvent): the queued event which failed
    """
    pass


from typing import List, Optional
from pydantic import BaseModel, Field

class FetcherConfig(BaseModel):
    """
    The configuration of a fetcher, used as part of a FetchEvent
    Fetch Provider's have their own uniqueue events and configurations.
    Configurations  
    """
    fetcher: Optional[str] = Field(None, description="indicates to OPAL client that it should use a custom FetcherProvider to fetch the data")

class FetchEvent(BaseModel):
    """
    Event used to describe an queue fetching tasks
    Design note -
        By using a Pydantic model - we can create a potentially transfer FetchEvents to be handled by other network nodes (perhaps via RPC)
    """
    # Event id to be filled by the engine
    id: str = None
    # optional name of the specific event
    name: str = None
    # A string identifying the fetcher class to use (as registered in the fetcher register)
    fetcher: str
    # The url the event targets for fetching
    url: str
    # Specific fetcher configuration (overridden by deriving event classes (FetcherConfig)
    config: dict = None
    # Tenacity.retry - Override default retry configuration for this event     
    retry: dict = None







from enum import Enum
from typing import Dict, Any

from jwt.algorithms import get_default_algorithms
from cryptography.hazmat._types import (
    _PRIVATE_KEY_TYPES,
    _PUBLIC_KEY_TYPES,
)

# custom types
PrivateKey = _PRIVATE_KEY_TYPES
PublicKey = _PUBLIC_KEY_TYPES
JWTClaims = Dict[str, Any]


class EncryptionKeyFormat(str, Enum):
    """
    represent the supported formats for storing encryption keys.

    - PEM (https://en.wikipedia.org/wiki/Privacy-Enhanced_Mail)
    - SSH (RFC4716) or short format (RFC4253, section-6.6, explained here: https://coolaj86.com/articles/the-ssh-public-key-format/)
    - DER (https://en.wikipedia.org/wiki/X.690#DER_encoding)
    """
    pem = 'pem'
    ssh = 'ssh'
    der = 'der'


# dynamic enum because pyjwt does not define one
# see: https://pyjwt.readthedocs.io/en/stable/algorithms.html for possible values
JWTAlgorithm = Enum('JWTAlgorithm', [(k,k) for k in get_default_algorithms().keys()])


"""
Simple HTTP get data fetcher using requests
supports 
"""

from enum import Enum
from typing import Any
from aiohttp import ClientResponse, ClientSession
from pydantic import validator
from ..fetch_provider import BaseFetchProvider
from ..events import FetcherConfig, FetchEvent
from ..logger import get_logger
from ...security.sslcontext import get_custom_ssl_context
from ...http import is_http_error_response

logger = get_logger("http_fetch_provider")


class HttpMethods(Enum):
    GET="get"
    POST="post"
    PUT="put"
    PATCH="patch"
    HEAD="head"
    DELETE="delete"
    
class HttpFetcherConfig(FetcherConfig):
    """
    Config for HttpFetchProvider's Adding HTTP headers 
    """
    headers: dict = None
    is_json: bool = True
    process_data: bool = True
    method:HttpMethods = HttpMethods.GET
    data:Any = None

    @validator('method')
    def force_enum(cls, v):
        if isinstance(v, str):
            return HttpMethods(v)
        if isinstance(v, HttpMethods):
            return v
        raise ValueError(f'invalid value: {v}')

    class Config:
        use_enum_values = True

class HttpFetchEvent(FetchEvent):
    fetcher: str = "HttpFetchProvider"
    config: HttpFetcherConfig = None

class HttpFetchProvider(BaseFetchProvider):

    def __init__(self, event: HttpFetchEvent) -> None:
        self._event: HttpFetchEvent
        if event.config is None:
            event.config = HttpFetcherConfig()
        super().__init__(event)
        self._session = None
        self._custom_ssl_context = get_custom_ssl_context()
        self._ssl_context_kwargs = {'ssl': self._custom_ssl_context} if self._custom_ssl_context is not None else {}

    def parse_event(self, event: FetchEvent) -> HttpFetchEvent:
        return HttpFetchEvent(**event.dict(exclude={"config"}), config=event.config)   

    async def __aenter__(self):
        headers = {}
        if self._event.config.headers is not None:
            headers = self._event.config.headers        
        self._session = await ClientSession(headers=headers).__aenter__()
        return self 

    async def __aexit__(self, exc_type=None, exc_val=None, tb=None):
        await self._session.__aexit__(exc_type=exc_type, exc_val=exc_val, exc_tb=tb)

    async def _fetch_(self):
        logger.debug(f"{self.__class__.__name__} fetching from {self._url}")
        http_method = self.match_http_method_from_type(self._session, self._event.config.method)
        if self._event.config.data is not None:
            result = await http_method(self._url, data=self._event.config.data, **self._ssl_context_kwargs)
        else:
            result = await http_method(self._url, **self._ssl_context_kwargs)
        return result

    @staticmethod
    def match_http_method_from_type(session:ClientSession, method_type:HttpMethods):
        return getattr(session, method_type.value)

    async def _process_(self, res: ClientResponse):
        # do not process data when the http response is an error
        if is_http_error_response(res):
            return res

        # if we are asked to process the data before we return it
        if self._event.config.process_data:
            # if data is JSON
            if self._event.config.is_json:
                data = await res.json()
            else:
                data = await res.text() 
            return data
        # return raw result
        else:
            return res



from typing import Optional

import jwt
from fastapi import HTTPException, status
from jwt.algorithms import Algorithm, get_default_algorithms
from jwt.api_jwk import PyJWK

from opal_common.authentication.types import JWTAlgorithm, JWTClaims, PublicKey
from opal_common.logger import logger


class Unauthorized(HTTPException):
    """
    HTTP 401 Unauthorized exception.
    """
    def __init__(self, description="Bearer token is not valid!", **kwargs):
        super().__init__(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail={
                "error": description,
                **kwargs
            },
            headers={"WWW-Authenticate": "Bearer"}
        )

class JWTVerifier:
    """
    given a cryptographic public key, can verify jwt tokens.
    """
    def __init__(
        self,
        public_key: Optional[PublicKey],
        algorithm: JWTAlgorithm,
        audience: str,
        issuer: str,
    ):
        """
        inits the signer if and only if the keys provided to __init__
        were generate together are are valid. otherwise will throw.

        JWT verifier can be initialized without a public key (None)
        in which case verifier.enabled == False and jwt verification is turned off.

        This allows opal to run both in secure mode (with jwt-based authentication)
        and in insecure mode (intended for development environments and running locally).

        Args:
            public_key (PublicKey): a valid public key or None
            algorithm (JWTAlgorithm): the jwt algorithm to use
                (possible values: https://pyjwt.readthedocs.io/en/stable/algorithms.html)
            audience (string): the value for the aud claim: https://tools.ietf.org/html/rfc7519#section-4.1.3
            issuer (string): the value for the iss claim: https://tools.ietf.org/html/rfc7519#section-4.1.1
        """
        self._public_key = public_key
        self._algorithm: str = algorithm.value
        self._audience = audience
        self._issuer = issuer
        self._enabled = True
        self._verify_public_key()

    def _verify_public_key(self):
        """
        verifies whether or not the public key is a valid crypto key
        (according to the JWT algorithm).
        """
        if self._public_key is not None:
            # save jwk
            try:
                self._jwk: PyJWK = PyJWK.from_json(self.get_jwk(), algorithm=self._algorithm)
            except jwt.exceptions.InvalidKeyError as e:
                logger.error(f"Invalid public key for jwt verification, error: {e}!")
                self._disable()
        else:
            self._disable()

    def get_jwk(self) -> str:
        """
        returns the jwk json contents
        """
        algorithm: Optional[Algorithm] = get_default_algorithms().get(self._algorithm)
        if algorithm is None:
            raise ValueError(f"invalid jwt algorithm: {self._algorithm}")
        return algorithm.to_jwk(self._public_key)

    def _disable(self):
        self._enabled = False

    @property
    def enabled(self):
        """
        whether or not the verifier has valid cryptographic keys
        """
        return self._enabled

    def verify(self, token: str) -> JWTClaims:
        """
        verifies a JWT token is valid.
        if valid returns dict with jwt claims, otherwise throws.
        """
        try:
            return jwt.decode(
                token,
                self._public_key,
                algorithms=[self._algorithm],
                audience=self._audience,
                issuer=self._issuer
            )
        except jwt.ExpiredSignatureError:
            raise Unauthorized(token=token, description="Access token is expired")
        except jwt.InvalidAudienceError:
            raise Unauthorized(token=token, description="Invalid access token: invalid audience claim")
        except jwt.InvalidIssuerError:
            raise Unauthorized(token=token, description="Invalid access token: invalid issuer claim")
        except jwt.DecodeError:
            raise Unauthorized(token=token, description="Could not decode access token")
        except Exception:
            raise Unauthorized(token=token, description="Unknown JWT error")



from uuid import UUID, uuid4
from enum import Enum
from datetime import datetime, timedelta
from typing import Optional

from pydantic import BaseModel, Field, validator


PEER_TYPE_DESCRIPTION = "The peer type we generate access token for, i.e: opal client, data provider, etc."
TTL_DESCRIPTION = "Token lifetime (timedelta), can accept duration in seconds or ISO_8601 format." + \
    " see: https://en.wikipedia.org/wiki/ISO_8601#Durations"
CLAIMS_DESCRIPTION = "extra claims to attach to the jwt"

class PeerType(str, Enum):
    client = 'client'
    datasource = 'datasource'
    listener = 'listener'

class AccessTokenRequest(BaseModel):
    """
    a request to generate an access token to opal server
    """
    id: UUID = Field(default_factory=uuid4)
    type: PeerType = Field(PeerType.client, description=PEER_TYPE_DESCRIPTION)
    ttl: timedelta = Field(timedelta(days=365), description=TTL_DESCRIPTION)
    claims: dict = Field({}, description=CLAIMS_DESCRIPTION)

    @validator('type')
    def force_enum(cls, v):
        if isinstance(v, str):
            return PeerType(v)
        if isinstance(v, PeerType):
            return v
        raise ValueError(f'invalid value: {v}')

    class Config:
        use_enum_values = True
        allow_population_by_field_name = True


class TokenDetails(BaseModel):
    id: UUID
    type: PeerType = Field(PeerType.client, description=PEER_TYPE_DESCRIPTION)
    expired: datetime
    claims: dict

class AccessToken(BaseModel):
    token: str
    type: str = "bearer"
    details: Optional[TokenDetails]


from pathlib import Path

def is_data_module(path: Path) -> bool:
    """
    Only json files named `data.json` can be included in offical OPA
    bundles as static data files. checks if a given path points to such file.
    """
    return path.name == 'data.json'

def is_rego_module(path: Path) -> bool:
    """
    Checks if a given path points to a rego file (extension == .rego).
    Only rego files are allowed in official OPA bundles as policy files.
    """
    return path.suffix == '.rego'


import asyncio
import glob
import hashlib
import os
import threading
import logging

from typing import Tuple, Coroutine


def get_filepaths_with_glob(root_path: str, file_regex: str):
    return glob.glob(os.path.join(root_path, file_regex))


def hash_file(tmp_file_path):
    BUF_SIZE = 65536  # lets read stuff in 64kb chunks!
    sha256_hash = hashlib.sha256()
    with open(tmp_file_path, "rb") as file:
        while True:
            data = file.read(BUF_SIZE)
            if not data:
                break
            sha256_hash.update(data)
    return sha256_hash.hexdigest()


def get_authorization_header(token: str) -> Tuple[str, str]:
    return ("Authorization", f"Bearer {token}")

def sorted_list_from_set(s: set) -> list:
    l = list(s)
    l.sort()
    return l

async def thread_worker(queue: asyncio.Queue, logger: logging.Logger):
    """
    The worker task is *running and then awaiting* a coroutine that was scheduled
    on the thread's async loop from *OUTSIDE* (i.e: from another thread).

    Args:
        queue (asyncio.Queue): The Queue
        engine (BaseFetchingEngine): The engine itself
    """
    while True:
        # get the next coroutine scheduled on the thread's queue
        # this may block until another coroutine is scheduled
        coro: Coroutine = await queue.get()

        try:
            # await on the coroutine and possibly block *this* worker
            await coro
        except Exception as err:
            logger.exception(f"Scheduled coroutine - {coro} failed")
        finally:
            # Notify the queue that the "work item" has been processed.
            queue.task_done()

class AsyncioEventLoopThread(threading.Thread):
    """
    This class enable a sync (or async) program to run (another)
    asyncio event loop in a separate thread without blocking the main
    thread or interfering with the main thread's asyncio loop if such exists.

    usage:
    t = AsyncioEventLoopThread()

    # not yet running
    t.create_task(coroutine1())
    t.create_task(coroutine2())

    # will start the event loop and all scheduled tasks
    t.start()
    """
    DEFAULT_WORKER_COUNT = 5

    def __init__(self, *args, loop=None, worker_count=DEFAULT_WORKER_COUNT, **kwargs):
        super().__init__(*args, **kwargs)
        self.daemon = True
        self.running = False
        self.loop = loop or asyncio.new_event_loop()
        # the thread is assigned a main logger bearing its name
        self.logger = logging.getLogger(self.name)
        # The internal task queue
        self._queue = asyncio.Queue(loop=self.loop)
        # Worker working the queue
        self._tasks = []

        # create worker tasks
        for _ in range(worker_count):
            self._create_worker()

    def run(self):
        """
        called by the default threading.Thread.start() method.
        runs the main activity of the thread, which in our case
        is simply running the asyncio loop until it stop.
        """
        self.running = True
        # does not return (thread will keep running) until loop.stop() is called
        if not self.loop.is_running():
            self.loop.run_forever()

    def stop(self):
        """
        Stops the thread.
        (Stop the async loop running on the thread and then joins the main thread).
        """
        self.run_coro(self._shutdown()) # will block until _shutdown() returns
        self.join() # will block until run() exits
        self.running = False

    def _create_worker(self) -> asyncio.Task:
        """
        Create an asyncio worker task to work the thread's queue.
        """
        task = self.loop.create_task(thread_worker(self._queue, self.logger))
        self._tasks.append(task)
        return task

    async def _shutdown(self):
        """
        Cancel and wait on the thread's async tasks
        """
        tasks = [
            t for t in asyncio.all_tasks(loop=self.loop)
            if t is not asyncio.current_task()
        ]
        for task in tasks:
            task.cancel()
        # Wait until all tasks are cancelled.
        await asyncio.gather(*tasks, return_exceptions=True)
        # stop the thread async loop
        self.loop.stop()

    def create_task(self, coro: Coroutine):
        """
        Creates a task on the thread's asyncio loop *without* waiting for it to finish.
        This is intended to be called from the parent thread as a set-and-forget.

        the scheduled coroutine is put on the thread's queue
        and is consumed by one of the thread workers.
        """
        async def _schedule_task():
            """
            since the queue is infinite, queue.put() will not block
            """
            await self._queue.put(coro)

        # the asyncio loop might not be running yet (if the thread was
        # not yet started), therefore we do not block on the result.
        return asyncio.run_coroutine_threadsafe(_schedule_task(), loop=self.loop)

    def run_coro(self, coro: Coroutine):
        """
        can be called from the main thread, but will run the coroutine
        on the event loop thread. the main thread will block until a
        result is returned. calling run_coro() is thread-safe.
        """
        return asyncio.run_coroutine_threadsafe(coro, loop=self.loop).result()






"""
Simple HTTP get data fetcher using requests
supports 
"""

import requests

from ..fetch_provider import BaseFetchProvider
from ..events import FetcherConfig, FetchEvent
from ..logger import get_logger

from fastapi_websocket_rpc.websocket_rpc_client import WebSocketRpcClient
from fastapi_websocket_rpc.rpc_methods import RpcMethodsBase

logger = get_logger("rpc_fetch_provider")


class FastApiRpcFetchConfig(FetcherConfig):
    """
    Config for FastApiRpcFetchConfig's Adding HTTP headers 
    """
    rpc_method_name: str
    rpc_arguments: dict


class FastApiRpcFetchEvent(FetchEvent):
    fetcher: str = "FastApiRpcFetchProvider"
    config: FastApiRpcFetchConfig


class FastApiRpcFetchProvider(BaseFetchProvider):

    def __init__(self, event: FastApiRpcFetchEvent) -> None:
        self._event: FastApiRpcFetchEvent
        super().__init__(event)

    def parse_event(self, event: FetchEvent) -> FastApiRpcFetchEvent:
        return FastApiRpcFetchEvent(**event.dict(exclude={"config"}), config=event.config)   


    async def _fetch_(self):
        assert self._event is not None, "FastApiRpcFetchEvent not provided for FastApiRpcFetchProvider"
        args = self._event.config.rpc_arguments
        method = self._event.config.rpc_method_name
        result = None
        logger.info(f"{self.__class__.__name__} fetching from {self._url} with RPC call {method}({args})")
        async with WebSocketRpcClient(self._url,
                                      # we don't expose anything to the server
                                      RpcMethodsBase(),
                                      default_response_timeout=4) as client:
            result = await client.call(method, args)
        return result



import os
import ssl

from typing import Optional

from opal_common.config import opal_common_config

def get_custom_ssl_context() -> Optional[ssl.SSLContext]:
    """
    Potentially (if enabled), returns a custom ssl context that respect self-signed certificates.
    More accurately, may return an ssl context that respects a local CA as a valid issuer.
    """
    if not opal_common_config.CLIENT_SELF_SIGNED_CERTIFICATES_ALLOWED:
        return None

    ca_file: Optional[str] = opal_common_config.CLIENT_SSL_CONTEXT_TRUSTED_CA_FILE

    if ca_file is None:
        return None

    if len(ca_file) == 0:
        return None

    ca_file_path = os.path.expanduser(ca_file)
    if not os.path.isfile(ca_file_path):
        return None

    return ssl.create_default_context(cafile=ca_file_path)


from logging import basicConfig
from opal_common.fetcher.providers.http_fetch_provider import HttpFetcherConfig
from typing import Dict, List, Optional, Tuple, Union
from pydantic import BaseModel, Field, root_validator, AnyHttpUrl

from opal_common.fetcher.events import FetcherConfig


class DataSourceEntry(BaseModel):
    """
    Data source configuration - where client's should retrive data from and how they should store it
    """
    # How to obtain the data
    url:str = Field(..., description="Url source to query for data")
    config:dict =  Field(None, description="Suggested fetcher configuration (e.g. auth or method) to fetch data with")
    # How to catalog data
    topics:List[str] = Field(None, description="topics the data applies to")
    # How to save the data
    # see https://www.openpolicyagent.org/docs/latest/rest-api/#data-api path is the path nested under <OPA_SERVER>/<version>/data
    dst_path:str = Field("", description="OPA data api path to store the document at")
    save_method:str = Field("PUT", description="Method used to write into OPA - PUT/PATCH")

class DataSourceConfig(BaseModel):
    """
    Static list of Data Source Entries returned to client.

    Answers this question for the client:
    from where should i get the full picture of data i need? (as opposed to incremental data updates)
    """
    entries: List[DataSourceEntry] = Field(..., description="list of data sources and how to fetch from them")

class ServerDataSourceConfig(BaseModel):
    """
    As its data source configuration, the server can either hold:

    1) A static DataSourceConfig returned to all clients regardless of identity.
    If all clients need the same config, this is the way to go.

    2) A redirect url (external_source_url), to which the opal client will be redirected when requesting
    its DataSourceConfig. The client will issue the same request (with the same headers, including the
    JWT token identifying it) to the url configured. This option is good if each client must receive a
    different base data configuration, for example for a multi-tenant deployment.

    By providing the server that serves external_source_url the value of OPAL_AUTH_PUBLIC_KEY, that server
    can validate the JWT and get it's claims, in order to apply authorization and/or other conditions before
    returning the data sources relevant to said client.
    """
    config: Optional[DataSourceConfig] = Field(None, description="static list of data sources and how to fetch from them")
    external_source_url: Optional[AnyHttpUrl] = Field(None, description="external url to serve data sources dynamically." + \
        " if set, the clients will be redirected to this url when requesting to fetch data sources.")

    @root_validator
    def check_passwords_match(cls, values):
        config, redirect_url = values.get('config'), values.get('external_source_url')
        if config is None and redirect_url is None:
            raise ValueError('you must provide one of these fields: config, external_source_url')
        if config is not None and redirect_url is not None:
            raise ValueError('you must provide ONLY ONE of these fields: config, external_source_url')
        return values


class CallbackEntry(BaseModel):
    """
    an entry in the callbacks register.
    this schema is used by the callbacks api
    """
    key: Optional[str] = Field(None, description="unique id to identify this callback (optional)")
    url: str = Field(..., description="http/https url to call back on update")
    config: Optional[HttpFetcherConfig] = Field(None, description="optional http config for the target url (i.e: http method, headers, etc)")

class UpdateCallback(BaseModel):
    """
    Configuration of callbacks upon completion of a FetchEvent
    Allows notifying other services on the update flow

    Each callback is either a URL (str) or a tuple of a url and HttpFetcherConfig defining how to approach the URL
    """
    callbacks: List[Union[str, Tuple[str, HttpFetcherConfig] ]]

class DataUpdate(BaseModel):
    """
    DataSources used as OPAL-server configuration
    Data update sent to clients
    """
    # a UUID to identify this update (used as part of an updates complition callback)
    id: Optional[str] = None
    entries: List[DataSourceEntry] = Field(..., description="list of related updates the OPAL client should perform")
    reason: str = Field(None, description="Reason for triggering the update")
    # Configuration for how to notify other services on the status of Update
    callback: UpdateCallback = UpdateCallback(callbacks=[])


class DataEntryReport(BaseModel):
    """
    A report of the processing of a single DataSourceEntry
    """
    entry: DataSourceEntry = Field(..., description="The entry that was processed")
    # Was the entry successfully fetched
    fetched: Optional[bool] = False
    # Was the entry successfully saved into the policy-data-store
    saved: Optional[bool] = False
    # Hash of the returned data
    hash: Optional[str] = None



class DataUpdateReport(BaseModel):
    # the UUID fo the update this report is for
    update_id: Optional[str] = None
    # Each DataSourceEntry and how it was processed
    reports: List[DataEntryReport]
    # in case this is a policy update, the new hash committed the policy store.
    policy_hash: Optional[str] = None






import fcntl
import time
import os
import asyncio

from typing import Optional
from opal_common.logger import logger

DEFAULT_LOCK_ATTEMPT_INTERVAL = 5.0

class NamedLock:
    """
    creates a a file-lock (can be a normal file or a named pipe / fifo),
    and exposes a context manager to try to acquire the lock asyncronously.
    """
    def __init__(self, path: str, attempt_interval: float = DEFAULT_LOCK_ATTEMPT_INTERVAL):
        self._lock_file: str = path
        self._lock_file_fd = None
        self._attempt_interval = attempt_interval

    async def __aenter__(self):
        """
        using the lock as a context manager will try to acquire
        the lock until successful (without timeout)
        """
        await self.acquire()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        """
        releases the lock when exiting the lock context
        """
        await self.release()

    async def acquire(self, timeout: Optional[int] = None):
        """
        tries to acquire the lock.

        if unsuccessful, will sleep and then try again after the attempt interval.
        an optional timeout can be provided to give up before acquiring the lock,
        in case we reach timeout, function throws TimeoutError.
        """
        logger.debug("[{pid}] trying to acquire lock (lock={lock})", pid=os.getpid(), lock=self._lock_file)
        start_time = time.time()
        while True:
            if self._acquire():
                logger.debug("[{pid}] lock acquired! (lock={lock})", pid=os.getpid(), lock=self._lock_file)
                break
            await asyncio.sleep(self._attempt_interval)
            # potentially give up due to timeout (if timeout is set)
            if timeout is not None and time.time() - start_time > timeout:
                raise TimeoutError("could not acquire lock")

    async def release(self):
        """
        releases the lock
        """
        logger.debug("[{pid}] releasing lock (lock={lock})", pid=os.getpid(), lock=self._lock_file)
        fd = self._lock_file_fd
        self._lock_file_fd = None
        fcntl.flock(fd, fcntl.LOCK_UN)
        os.close(fd)

    @property
    def is_locked(self):
        """
        True, if the object holds the file lock.
        """
        return self._lock_file_fd is not None

    def _acquire(self) -> bool:
        """
        tries to acquire the lock, returns immediately regardless of success.
        returns True if lock was acquired successfully, False otherwise.
        """
        fd = os.open(self._lock_file, os.O_RDWR | os.O_CREAT | os.O_TRUNC)

        # try to acquire the lock, returns immediately
        try:
            fcntl.flock(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except (IOError, OSError):
            os.close(fd)
        else:
            self._lock_file_fd = fd
        return self.is_locked
